from .version import __version__

import torch
from memtorch.bh import *
from memtorch.bh.nonideality import *
from memtorch.map import *
from memtorch.mn import *
from memtorch.submodules import *
from memtorch.utils import *



import copy
import math
import random
import sys

import numpy as np
import pytest
import torch

import memtorch
from memtorch.bh.crossbar.Crossbar import simulate_matmul
from memtorch.bh.crossbar.Program import gen_programming_signal, naive_program
from memtorch.map.Parameter import naive_map


@pytest.mark.filterwarnings("ignore::Warning")
@pytest.mark.parametrize("shape", [(2, 2)])
def test_crossbar(shape):
    device = torch.device("cpu" if "cpu" in memtorch.__version__ else "cuda")
    memristor_model = memtorch.bh.memristor.LinearIonDrift
    memristor_model_params = {"time_series_resolution": 1e-3}
    pos_crossbar = memtorch.bh.crossbar.Crossbar(
        memristor_model, memristor_model_params, shape
    )
    pos_crossbar.devices[0][0].g = 1 / pos_crossbar.r_off_mean
    pos_crossbar.devices[0][1].g = 1 / pos_crossbar.r_on_mean
    neg_crossbar = copy.deepcopy(pos_crossbar)
    pos_conductance_matrix, neg_conductance_matrix = naive_map(
        torch.zeros(shape).uniform_(0, 1),
        memristor_model().r_on,
        memristor_model().r_off,
        memtorch.bh.crossbar.Scheme.DoubleColumn,
    )
    pos_crossbar.write_conductance_matrix(
        pos_conductance_matrix, transistor=False, programming_routine=naive_program
    )
    neg_crossbar.write_conductance_matrix(
        pos_conductance_matrix, transistor=False, programming_routine=naive_program
    )

    crossbar = memtorch.bh.crossbar.Crossbar(
        memristor_model, memristor_model_params, shape
    )
    conductance_matrix = naive_map(
        torch.zeros(shape).uniform_(0, 1),
        memristor_model().r_on,
        memristor_model().r_off,
        memtorch.bh.crossbar.Scheme.SingleColumn,
    )
    crossbar.write_conductance_matrix(conductance_matrix)
    if sys.version_info > (3, 6):
        crossbar.update(from_devices=False, parallelize=True)
        assert torch.all(
            torch.isclose(
                conductance_matrix.T[:, :],
                crossbar.conductance_matrix.cpu()[:, :],
                atol=1e-5,
            )
        )
        assert crossbar.devices[0][0].g == crossbar.conductance_matrix[0][0].item()

    crossbar.update(from_devices=False, parallelize=False)
    assert crossbar.devices[0][0].g == crossbar.conductance_matrix[0][0].item()
    inputs = torch.zeros(shape).uniform_(0, 1)
    assert torch.all(
        torch.isclose(
            simulate_matmul(inputs, crossbar).float(),
            torch.matmul(inputs, conductance_matrix.T).float().to(device),
            rtol=1e-1,
        )
    )
    programming_signal = gen_programming_signal(
        1, 1e-3, 0, 1, memristor_model_params["time_series_resolution"]
    )
    assert type(programming_signal) == tuple
    with pytest.raises(AssertionError):
        gen_programming_signal(
            1, 1e-4, 0, 1, memristor_model_params["time_series_resolution"]
        )

    point = (0, 0)
    row, column = point
    conductance_to_program = random.uniform(
        1 / crossbar.devices[row][column].r_off, 1 / crossbar.devices[row][column].r_on
    )
    crossbar.devices = naive_program(
        crossbar, (row, column), conductance_to_program, rel_tol=0.01
    )
    assert math.isclose(
        conductance_to_program, crossbar.devices[row][column].g, abs_tol=1e-4
    )
    with pytest.raises(AssertionError):
        naive_program(crossbar, (row, column), -1)



import math

import numpy as np

import memtorch
from memtorch.utils import clip

from .Memristor import Memristor as Memristor


class Data_Driven2021(Memristor):
    """An updated Data-Driven Verilog-A ReRAM Model. Based on the model used in the article at the following address:
    https://arxiv.org/abs/2012.02267. The main difference with the 2018 version (simply Data_Driven.py) is that it is
    less computationally demanding and achieves similar results. The creator of the data driven model as decided to
    return to the 2017 version for this reason in his recent works.
    The default parameters were determined experimentally.

    Parameters
    ----------
    time_series_resolution : float
        Time series resolution (s).
    r_off : float
        Off (maximum) resistance of the device (ohms).
    r_on : float
        On (minimum) resistance of the device (ohms).
    A_p : float
        A_p model parameter.
    A_n : float
        A_n model parameter.
    t_p : float
        t_p model parameter.
    t_n : float
        t_n model parameter.
    k_p : float
        k_p model parameter.
    k_n : float
        k_n model parameter.
    r_p : float
        r_p voltage-dependent resistive boundary function coefficients.
    r_n : float
        r_n voltage-dependent resistive boundary function coefficients.
    a_p : float
        a_p model parameter.
    a_n : float
        a_n model parameter.
    b_p : float
        b_p model parameter.
    b_n : float
        b_n model parameter.
    """

    def __init__(
        self,
        time_series_resolution=1e-10,
        r_off=3000,
        r_on=1600,
        A_p=600.10075,
        A_n=-34.5988399,
        t_p=-0.0212028,
        t_n=-0.05343997,
        k_p=5.11e-4,
        k_n=1.17e-3,
        r_p=[2699.2336, -672.930205],  # a0_p, a1_p
        r_n=[649.413746, -1474.32358],  # a0_n, a1_n
        a_p=0.32046175,
        a_n=0.32046175,
        b_p=2.71689828,
        b_n=2.71689828,
        **kwargs
    ):

        args = memtorch.bh.unpack_parameters(locals())
        super(Data_Driven2021, self).__init__(
            args.r_off, args.r_on, args.time_series_resolution, 0, 0
        )
        self.A_p = args.A_p
        self.A_n = args.A_n
        self.t_p = args.t_p
        self.t_n = args.t_n
        self.k_p = args.k_p
        self.k_n = args.k_n
        self.r_p = args.r_p
        self.r_n = args.r_n
        self.a_p = args.a_p
        self.a_n = args.a_n
        self.b_p = args.b_p
        self.b_n = args.b_n
        self.g = 1 / args.r_on

    def simulate(self, voltage_signal, return_current=False):
        len_voltage_signal = 1
        try:
            len_voltage_signal = len(voltage_signal)
        except:
            voltage_signal = [voltage_signal]

        if return_current:
            current = np.zeros(len_voltage_signal)

        np.seterr(all="raise")
        for t in range(0, len_voltage_signal):
            current_ = self.current(voltage_signal[t])
            self.g = 1 / self.resistance(voltage_signal[t])
            if return_current:
                current[t] = current_
        if return_current:
            return current

    def r_pn(self, voltage, a0, a1):
        """Function to return rp(v) or rn(v)
        From the 2017/2021 paper model calculations

        Parameters
        ----------
        voltage : float
            The current applied voltage (V).
        a0, a1: float
            The value of a0 and a1

        Returns
        -------
        float
        The rp or rn resistance (Ω).
        """
        return a0 + a1 * voltage

    def s_pn(self, voltage, A, t):
        """Function to return sp(v) or sn(v)
        From the 2017/2021 paper model calculations

        Parameters
        ----------
        voltage : float
        The current applied voltage (V).
        A, t: float
        The value of model params A (A_p or A_n) or t(t_p or t_n)

        Returns
        -------
        float
        The sp or sn variability.
        """
        return A * (math.exp(abs(voltage) / t) - 1)

    def dRdt(self, voltage):
        """Function to return dR/dT
        From the 2017/2021 paper model calculations

        Parameters
        ----------
        voltage : float
        The current applied voltage (V).
        a0, a1: float
        The value of a0 or a1

        Returns
        -------
        float
        The derivative with respect to time of the resistance
        """

        R = 1 / self.g
        if voltage > 0:
            r_p = self.r_pn(voltage, self.r_p[0], self.r_p[1])
            s_p = self.s_pn(voltage, self.A_p, self.t_p)
            return s_p * (r_p - R) ** 2
        if voltage <= 0:
            r_n = self.r_pn(voltage, self.r_n[0], self.r_n[1])
            s_n = self.s_pn(voltage, self.A_n, self.t_n)
            return s_n * (R - r_n) ** 2

        return

    def current(self, voltage):
        """Method to determine the current of the model given an applied voltage.

        Parameters
        ----------
        voltage : float
            The current applied voltage (V).

        Returns
        -------
        float
            The observed current (A).
        """
        if voltage > 0:
            return self.a_p * self.g * math.sinh(self.b_p * voltage)
        else:
            return self.a_n * self.g * math.sinh(self.b_n * voltage)

    def resistance(self, voltage):
        """Method to determine the resistance of the model given an applied voltage.
        Using the 2017/2021 model

        Parameters
        ----------
        voltage : float
            The current applied voltage (V).

        Returns
        -------
        float
            The observed resistance (Ω).
        """
        R0 = 1 / self.g
        if voltage > 0:
            r_p = self.r_pn(voltage, self.r_p[0], self.r_p[1])
            s_p = self.s_pn(voltage, self.A_p, self.t_p)
            resistance_ = (
                R0 + (s_p * r_p * (r_p - R0)) * self.time_series_resolution
            ) / (1 + s_p * (r_p - R0) * self.time_series_resolution)
            if resistance_ < r_p:
                return R0
            else:
                return max(
                    min(resistance_, self.r_off), self.r_on
                )  # Artificially confine the resistance between r_on and r_off

        elif voltage < 0:
            r_n = self.r_pn(voltage, self.r_n[0], self.r_n[1])
            s_n = self.s_pn(voltage, self.A_n, self.t_n)
            resistance_ = (
                R0 + (s_n * r_n * (r_n - R0)) * self.time_series_resolution
            ) / (1 + s_n * (r_n - R0) * self.time_series_resolution)
            if resistance_ > r_n:
                return R0
            else:
                return max(
                    min(resistance_, self.r_off), self.r_on
                )  # Artificially confine the resistance between r_on and r_off

        else:
            return 1 / self.g

    def set_conductance(self, conductance):
        conductance = clip(conductance, 1 / self.r_off, 1 / self.r_on)
        self.g = conductance

    def plot_hysteresis_loop(
        self,
        duration=1e-3,
        voltage_signal_amplitude=1.5,
        voltage_signal_frequency=10e3,
        return_result=False,
    ):
        return super().plot_hysteresis_loop(
            self,
            duration=duration,
            voltage_signal_amplitude=voltage_signal_amplitude,
            voltage_signal_frequency=voltage_signal_frequency,
            return_result=return_result,
        )

    def plot_bipolar_switching_behaviour(
        self,
        voltage_signal_amplitude=1.5,
        voltage_signal_frequency=10e3,
        log_scale=True,
        return_result=False,
    ):
        return super().plot_bipolar_switching_behaviour(
            self,
            voltage_signal_amplitude=voltage_signal_amplitude,
            voltage_signal_frequency=voltage_signal_frequency,
            log_scale=log_scale,
            return_result=return_result,
        )



import math
import warnings

import numpy as np
import torch
import torch.nn as nn

import memtorch
from memtorch.bh.crossbar.Crossbar import init_crossbar, simulate_matmul
from memtorch.bh.crossbar.Tile import tiled_inference
from memtorch.map.Input import naive_scale
from memtorch.map.Module import naive_tune
from memtorch.map.Parameter import naive_map


class Conv3d(nn.Conv3d):
    """nn.Conv3d equivalent.

    Parameters
    ----------
    convolutional_layer : torch.nn.Conv3d
        Convolutional layer to patch.
    memristor_model : memtorch.bh.memristor.Memristor.Memristor
        Memristor model.
    memristor_model_params : **kwargs
        Memristor model keyword arguments.
    mapping_routine : function
        Mapping routine to use.
    transistor : bool
        Used to determine if a 1T1R (True) or 1R arrangement (False) is simulated.
    programming_routine : function
        Programming routine to use.
    programming_routine_params : **kwargs
        Programming routine keyword arguments.
    p_l: float
        If not None, the proportion of weights to retain.
    scheme : memtorch.bh.Scheme
        Weight representation scheme.
    tile_shape : (int, int)
        Tile shape to use to store weights. If None, modular tiles are not used.
    max_input_voltage : float
        Maximum input voltage used to encode inputs. If None, inputs are unbounded.
    scaling_routine : function
        Scaling routine to use in order to scale batch inputs.
    scaling_routine_params : **kwargs
        Scaling routine keyword arguments.
    source_resistance : float
        The resistance between word/bit line voltage sources and crossbar(s).
    line_resistance : float
        The interconnect line resistance between adjacent cells.
    ADC_resolution : int
        ADC resolution (bit width). If None, quantization noise is not accounted for.
    ADC_overflow_rate : float
        Overflow rate threshold for linear quanitzation (if ADC_resolution is not None).
    quant_method: string
        Quantization method. Must be in ['linear', 'log', 'log_minmax', 'minmax', 'tanh'], or None.
    use_bindings : bool
        Used to determine if C++/CUDA bindings are used (True) or not (False).
    random_crossbar_init: bool
        Determines if the crossbar is to be initialized at random values in between Ron and Roff
    verbose : bool
        Used to determine if verbose output is enabled (True) or disabled (False).
    """

    def __init__(
        self,
        convolutional_layer,
        memristor_model,
        memristor_model_params,
        mapping_routine=naive_map,
        transistor=True,
        programming_routine=None,
        programming_routine_params={},
        p_l=None,
        scheme=memtorch.bh.Scheme.DoubleColumn,
        tile_shape=None,
        max_input_voltage=None,
        scaling_routine=naive_scale,
        scaling_routine_params={},
        source_resistance=None,
        line_resistance=None,
        ADC_resolution=None,
        ADC_overflow_rate=0.0,
        quant_method=None,
        use_bindings=True,
        random_crossbar_init=False,
        verbose=True,
        *args,
        **kwargs
    ):
        assert isinstance(
            convolutional_layer, nn.Conv3d
        ), "convolutional_layer is not an instance of nn.Conv3d."
        assert (
            convolutional_layer.groups != 2
        ), "groups=2 is not currently supported for convolutional layers."
        self.device = torch.device("cpu" if "cpu" in memtorch.__version__ else "cuda")
        self.transistor = transistor
        self.scheme = scheme
        self.tile_shape = tile_shape
        self.max_input_voltage = max_input_voltage
        self.scaling_routine = scaling_routine
        self.scaling_routine_params = scaling_routine_params
        self.source_resistance = source_resistance
        self.line_resistance = line_resistance
        self.ADC_resolution = ADC_resolution
        self.ADC_overflow_rate = ADC_overflow_rate
        if "cpu" not in memtorch.__version__:
            self.cuda_malloc_heap_size = 50
        else:
            self.cuda_malloc_heap_size = None

        if not transistor:
            assert (
                source_resistance is not None and source_resistance >= 0.0
            ), "Source resistance is invalid."
            assert (
                line_resistance is not None and line_resistance >= 0.0
            ), "Line resistance is invalid."

        if quant_method in memtorch.bh.Quantize.quant_methods:
            self.quant_method = quant_method
        else:
            self.quant_method = None

        if quant_method is not None:
            assert (
                ADC_resolution is not None
                and type(ADC_resolution) == int
                and ADC_resolution > 0
            ), "ADC resolution is invalid."
            assert (
                ADC_overflow_rate is not None
            ), "ADC_overflow_rate must be specified if quant_method is not None."

        self.use_bindings = use_bindings
        self.verbose = verbose
        self.forward_legacy_enabled = True
        super(Conv3d, self).__init__(
            convolutional_layer.in_channels,
            convolutional_layer.out_channels,
            convolutional_layer.kernel_size,
            stride=convolutional_layer.stride,
            padding=convolutional_layer.padding,
            dilation=convolutional_layer.dilation,
            groups=convolutional_layer.groups,
            **kwargs
        )
        self.weight.data = convolutional_layer.weight.data
        if convolutional_layer.bias is not None:
            self.bias.data = convolutional_layer.bias.data

        self.zero_grad()
        self.weight.requires_grad = False
        if convolutional_layer.bias is not None:
            self.bias.requires_grad = False

        self.crossbars, self.crossbar_operation = init_crossbar(
            weights=self.weight,
            memristor_model=memristor_model,
            memristor_model_params=memristor_model_params,
            transistor=transistor,
            mapping_routine=mapping_routine,
            programming_routine=programming_routine,
            programming_routine_params=programming_routine_params,
            p_l=p_l,
            scheme=scheme,
            tile_shape=tile_shape,
            use_bindings=use_bindings,
            cuda_malloc_heap_size=self.cuda_malloc_heap_size,
            random_crossbar_init=random_crossbar_init,
        )
        self.transform_output = lambda x: x
        if verbose:
            print("Patched %s -> %s" % (convolutional_layer, self))

    def forward(self, input):
        """Method to perform forward propagations.

        Parameters
        ----------
        input : torch.Tensor
            Input tensor.

        Returns
        -------
        torch.Tensor
            Output tensor.
        """
        if self.forward_legacy_enabled:
            return torch.nn.functional.conv3d(
                input.to(self.device),
                self.weight.to(self.device),
                bias=self.bias,
                stride=self.stride,
                padding=self.padding,
            )
        else:
            output_dim = [0, 0, 0]
            output_dim[0] = (
                int(
                    (input.shape[2] - self.kernel_size[0] + 2 * self.padding[0])
                    / self.stride[0]
                )
                + 1
            )
            output_dim[1] = (
                int(
                    (input.shape[3] - self.kernel_size[1] + 2 * self.padding[1])
                    / self.stride[1]
                )
                + 1
            )
            output_dim[2] = (
                int(
                    (input.shape[4] - self.kernel_size[2] + 2 * self.padding[2])
                    / self.stride[2]
                )
                + 1
            )
            out = torch.zeros(
                (
                    input.shape[0],
                    self.out_channels,
                    output_dim[0],
                    output_dim[1],
                    output_dim[2],
                )
            ).to(self.device)
            for batch in range(input.shape[0]):
                if not all(item == 0 for item in self.padding):
                    batch_input = nn.functional.pad(
                        input[batch],
                        pad=(
                            self.padding[2],
                            self.padding[2],
                            self.padding[1],
                            self.padding[1],
                            self.padding[0],
                            self.padding[0],
                        ),
                    )
                else:
                    batch_input = input[batch]

                batch_input = self.scaling_routine(
                    self, batch_input, **self.scaling_routine_params
                )
                unfolded_batch_input = (
                    batch_input.unfold(1, self.kernel_size[0], self.stride[0])
                    .unfold(2, self.kernel_size[1], self.stride[1])
                    .unfold(3, self.kernel_size[2], self.stride[2])
                    .permute(1, 2, 3, 0, 4, 5, 6)
                    .reshape(
                        -1,
                        (self.in_channels // self.groups)
                        * self.kernel_size[0]
                        * self.kernel_size[1]
                        * self.kernel_size[2],
                    )
                )
                if hasattr(self, "non_linear"):
                    warnings.warn(
                        "Non-liner modeling does not currently account for source and line resistances."
                    )
                    if self.tile_shape is not None:
                        tiles_map = self.crossbars[0].tiles_map
                        crossbar_shape = (
                            self.crossbars[0].rows,
                            self.crossbars[0].columns,
                        )
                    else:
                        tiles_map = None
                        crossbar_shape = None

                    if hasattr(self, "simulate"):
                        nl = False
                    else:
                        nl = True

                    out_ = (
                        self.crossbar_operation(
                            self.crossbars,
                            lambda crossbar, input_: simulate_matmul(
                                unfolded_batch_input,
                                crossbar,
                                nl=nl,
                                tiles_map=tiles_map,
                                crossbar_shape=crossbar_shape,
                                max_input_voltage=self.max_input_voltage,
                                ADC_resolution=self.ADC_resolution,
                                ADC_overflow_rate=self.ADC_overflow_rate,
                                quant_method=self.quant_method,
                                use_bindings=self.use_bindings,
                            ),
                            input_=unfolded_batch_input,
                        )
                        .to(self.device)
                        .T
                    )
                else:
                    if self.tile_shape is not None:
                        out_ = tiled_inference(
                            unfolded_batch_input, self, transistor=self.transistor
                        ).T
                    else:
                        devices = self.crossbar_operation(
                            self.crossbars,
                            lambda crossbar: crossbar.conductance_matrix,
                        )
                        if self.transistor:
                            out_ = torch.matmul(
                                unfolded_batch_input,
                                devices,
                            ).T
                        else:
                            out_ = memtorch.bh.crossbar.Passive.solve_passive(
                                devices,
                                unfolded_batch_input.to(self.device),
                                torch.zeros(
                                    unfolded_batch_input.shape[0], devices.shape[1]
                                ),
                                self.source_resistance,
                                self.line_resistance,
                                n_input_batches=unfolded_batch_input.shape[0],
                                use_bindings=self.use_bindings,
                                cuda_malloc_heap_size=self.cuda_malloc_heap_size,
                            ).T

                        if self.quant_method is not None:
                            out_ = memtorch.bh.Quantize.quantize(
                                out_,
                                quant=self.ADC_resolution,
                                overflow_rate=self.ADC_overflow_rate,
                                quant_method=self.quant_method,
                            )

                    out[batch] = out_.view(size=(1, self.out_channels, *output_dim))

            out = self.transform_output(out)
            if not self.bias is None:
                out[batch] += self.bias.data.view(-1, 1, 1, 1).expand_as(out[batch])

            return out

    def tune(self, input_batch_size=4, input_shape=32):
        """Tuning method."""
        self.transform_output = naive_tune(
            self,
            (
                input_batch_size,
                (self.in_channels // self.groups),
                input_shape,
                input_shape,
                input_shape,
            ),
            self.verbose,
        )

    def __str__(self):
        return (
            "bh.Conv3d(in_channels=%d, out_channels=%d, kernel_size=(%d, %d, %d), stride=(%d, %d, %d), padding=(%d, %d, %d))"
            % (
                self.in_channels,
                self.out_channels,
                self.kernel_size[0],
                self.kernel_size[1],
                self.kernel_size[2],
                self.stride[0],
                self.stride[1],
                self.stride[2],
                self.padding[0],
                self.padding[1],
                self.padding[2],
            )
        )



import math

import numpy as np
import torch

import memtorch


def Jogelkar(x, p=1):
    """Jogelkar window function.

    Parameters
    ----------
    x : float
        State variable.
    p : int
        p constant.
    """
    return float(1 - (2 * x - 1)) ** (2 * p)



import codecs
import gzip
import lzma
import math
import os
import os.path
import string
import warnings
from typing import IO, Any, Callable, Dict, List, Optional, Tuple, Union
from urllib.error import URLError

import numpy as np
import torch
import torchvision
from PIL import Image
from torchvision import datasets, transforms
from torchvision.datasets.utils import (
    download_and_extract_archive,
    download_url,
    extract_archive,
    verify_str_arg,
)

import memtorch


def convert_range(old_value, old_min, old_max, new_min, new_max):
    """Method to convert values between two ranges.

    Parameters
    ----------
    old_value : object
        Old value(s) to convert. May be a single number, vector or tensor.
    old_min : float
        Minimum old value.
    old_max : float
        Maximum old value.
    new_min : float
        Minimum new value.
    new_max : float
        Maximum new value.

    Returns
    -------
    object
        New value(s).
    """
    return (
        ((old_value - old_min) * (new_max - new_min)) / (old_max - old_min)
    ) + new_min


def clip(value, lower, upper):
    """Method to clip a float between lower and upper bounds.

    Parameters
    ----------
    value : float
        Value to clip.
    lower : float
        Lower bound.
    upper : float
        Upper bound.

    Returns
    -------
    float
        Clipped float.
    """
    return lower if value < lower else upper if value > upper else value


def pad_tensor(tensor, tile_shape):
    """Method to zero-pad a tensor.

    Parameters
    ----------
    tensor : torch.tensor
        Tensor to zero-pad.
    tile_shape : (int, int)
        Tile shape to pad tensor for.

    Returns
    -------
    torch.tensor
        Padded tensor.
    """
    assert (
        len(tensor.shape) == 1 or len(tensor.shape) == 2
    ), "tensor.shape must be 1 or 2 dimensional."
    if len(tensor.shape) == 1:
        tensor_padded = torch.zeros(
            (math.ceil(tensor.shape[0] / tile_shape[0]) * tile_shape[0])
        )
        tensor_padded[0 : tensor.shape[0]] = tensor
    elif len(tensor.shape) == 2:
        tensor_padded = torch.zeros(
            (
                math.ceil(tensor.shape[0] / tile_shape[0]) * tile_shape[0],
                math.ceil(tensor.shape[1] / tile_shape[1]) * tile_shape[1],
            )
        )
        tensor_padded[0 : tensor.shape[0], 0 : tensor.shape[1]] = tensor

    return tensor_padded


def LoadMNIST(batch_size=32, validation=True, num_workers=1):
    """Method to load the MNIST dataset.

    Parameters
    ----------
    batch_size : int
        Batch size.
    validation : bool
        Load the validation set (True).
    num_workers : int
        Number of workers to use.

    Returns
    -------
    list of torch.utils.data
        The train, validiation, and test loaders.
    """
    root = "data"
    transform = transforms.Compose([transforms.ToTensor()])
    full_train_set = torchvision.datasets.MNIST(
        root=root, train=True, transform=transform, download=True
    )
    test_set = torchvision.datasets.MNIST(
        root=root, train=False, transform=transform, download=True
    )
    if validation:
        train_size = int(0.8 * len(full_train_set))
        validation_size = len(full_train_set) - train_size
        train_set, validation_set = torch.utils.data.random_split(
            full_train_set, [train_size, validation_size]
        )
        train_loader = torch.utils.data.DataLoader(
            train_set, batch_size=batch_size, shuffle=True, num_workers=num_workers
        )
        validation_loader = torch.utils.data.DataLoader(
            validation_set, batch_size=batch_size, shuffle=True, num_workers=num_workers
        )
    else:
        train_loader = torch.utils.data.DataLoader(
            full_train_set, batch_size=batch_size, shuffle=True, num_workers=num_workers
        )
        validation_loader = None

    test_loader = torch.utils.data.DataLoader(
        test_set, batch_size=int(batch_size / 2), shuffle=False, num_workers=num_workers
    )
    return train_loader, validation_loader, test_loader


def LoadCIFAR10(batch_size=32, validation=True, num_workers=1):
    """Method to load the CIFAR-10 dataset.

    Parameters
    ----------
    batch_size : int
        Batch size.
    validation : bool
        Load the validation set (True).
    num_workers : int
        Number of workers to use.

    Returns
    -------
    list of torch.utils.data
        The train, validiation, and test loaders.
    """
    root = "data"
    transform = transforms.Compose([transforms.ToTensor()])
    full_train_set = torchvision.datasets.CIFAR10(
        root=root, train=True, download=True, transform=transform
    )
    test_set = torchvision.datasets.CIFAR10(
        root=root, train=False, download=True, transform=transform
    )
    if validation:
        train_size = int(0.8 * len(full_train_set))
        validation_size = len(full_train_set) - train_size
        train_set, validation_set = torch.utils.data.random_split(
            full_train_set, [train_size, validation_size]
        )
        train_loader = torch.utils.data.DataLoader(
            train_set, batch_size=batch_size, shuffle=True, num_workers=num_workers
        )
        validation_loader = torch.utils.data.DataLoader(
            validation_set, batch_size=batch_size, shuffle=True, num_workers=num_workers
        )
    else:
        train_loader = torch.utils.data.DataLoader(
            full_train_set, batch_size=batch_size, shuffle=True, num_workers=num_workers
        )
        validation_loader = None

    test_loader = torch.utils.data.DataLoader(
        test_set, batch_size=int(batch_size / 2), shuffle=False, num_workers=num_workers
    )
    return train_loader, validation_loader, test_loader



import inspect

import pytest
import torch

import memtorch
from memtorch.bh.crossbar.Program import naive_program
from memtorch.map.Parameter import naive_map
from memtorch.mn.Module import patch_model, supported_module_parameters


@pytest.fixture
def debug_networks():
    default_kwargs = {
        "in_features": 2,
        "out_features": 2,
        "in_channels": 1,
        "out_channels": 2,
        "kernel_size": 1,
        "padding": 1,
        "bias": True,
    }
    networks = []
    device = torch.device("cpu" if "cpu" in memtorch.__version__ else "cuda")
    for supported_module_parameter in supported_module_parameters:

        class Network(torch.nn.Module):
            def __init__(self):
                super(Network, self).__init__()
                layer_type = supported_module_parameters[
                    supported_module_parameter
                ].__bases__[0]
                layer_args = list(inspect.signature(layer_type.__init__).parameters)
                args = {}
                layer_args.pop(0)
                for layer_arg in layer_args:
                    if layer_arg in default_kwargs:
                        args[layer_arg] = default_kwargs[layer_arg]

                self.layer = layer_type(**args)

            def forward(self, input):
                return self.layer(input)

        networks.append(Network().to(device))

    return networks


@pytest.fixture
def debug_patched_networks(debug_networks):
    def debug_patched_networks_(tile_shape, quant_method):
        if quant_method is not None:
            ADC_resolution = 8
        else:
            ADC_resolution = None

        networks = debug_networks
        device = torch.device("cpu" if "cpu" in memtorch.__version__ else "cuda")
        patched_networks = []
        for network in networks:
            patched_networks.append(
                patch_model(
                    network,
                    memristor_model=memtorch.bh.memristor.LinearIonDrift,
                    memristor_model_params={"time_series_resolution": 0.1},
                    module_parameters_to_patch=[type(network.layer)],
                    mapping_routine=naive_map,
                    transistor=True,
                    programming_routine=None,
                    scheme=memtorch.bh.Scheme.SingleColumn,
                    tile_shape=tile_shape,
                    max_input_voltage=1.0,
                    ADC_resolution=ADC_resolution,
                    quant_method=quant_method,
                )
            )

        return patched_networks

    return debug_patched_networks_



import copy

import numpy as np
import torch
from numpy.core.numeric import cross

import memtorch


def apply_finite_conductance_states(layer, n_conductance_states):
    """Method to model a finite number of conductance states for devices within a memristive layer.

    Parameters
    ----------
    layer : memtorch.mn
        A memrstive layer.
    n_conductance_states : int
        Number of finite conductance states to model.

    Returns
    -------
    memtorch.mn
        The patched memristive layer.
    """
    device = torch.device("cpu" if "cpu" in memtorch.__version__ else "cuda")
    assert (
        int(n_conductance_states) == n_conductance_states
    ), "n_conductance_states must be a whole number."

    def apply_finite_conductance_states_to_crossbar(crossbar, n_conductance_states):
        crossbar.update()
        conductance_matrix_ = copy.deepcopy(crossbar.conductance_matrix)
        try:
            r_on = np.nan_to_num(
                np.array(np.vectorize(lambda x: x.r_on)(crossbar.devices)),
                copy=False,
                nan=crossbar.r_on_mean,
                posinf=crossbar.r_on_mean,
                neginf=crossbar.r_on_mean,
            )
            r_on[r_on == 0] = crossbar.r_on_mean
            r_off = np.nan_to_num(
                np.array(np.vectorize(lambda x: x.r_off)(crossbar.devices)),
                copy=False,
                nan=crossbar.r_off_mean,
                posinf=crossbar.r_off_mean,
                neginf=crossbar.r_off_mean,
            )
            r_off[r_off == 0] = crossbar.r_off_mean
            if np.unique(r_on).size == 1:
                r_on = torch.ones(
                    crossbar.conductance_matrix.shape, device=device
                ).float() * float(crossbar.r_on_mean)
            else:
                r_on = torch.from_numpy(r_on).float().cuda()

            if np.unique(r_off).size == 1:
                r_off = torch.ones(
                    crossbar.conductance_matrix.shape, device=device
                ).float() * float(crossbar.r_off_mean)
            else:
                r_off = torch.from_numpy(r_off).float().cuda()

            conductance_matrix_shape = crossbar.conductance_matrix.shape
            crossbar.conductance_matrix = (
                memtorch.bh.Quantize.quantize(
                    crossbar.conductance_matrix.view(-1),
                    n_conductance_states,
                    min=1 / r_off.view(-1),
                    max=1 / r_on.view(-1),
                    override_original=True,
                )
                .view(conductance_matrix_shape)
                .float()
            )
        except:
            crossbar.conductance_matrix = conductance_matrix_.float()

        return crossbar

    for i in range(len(layer.crossbars)):
        layer.crossbars[i] = apply_finite_conductance_states_to_crossbar(
            layer.crossbars[i], n_conductance_states
        )

    return layer



import sys

import pytest

sys.path.insert(0, "..")


def test_CUDA_flag():
    CUDA_is_false = False
    f = open("setup.py", "r")
    for line in f.readlines():
        if line.strip() == "CUDA = False":
            CUDA_is_false = True
            break

    assert CUDA_is_false



import math
import time
import warnings

import numpy as np
import torch

import memtorch


def naive_program(
    crossbar,
    point,
    conductance,
    rel_tol=0.01,
    pulse_duration=1e-3,
    refactory_period=0,
    pos_voltage_level=1.0,
    neg_voltage_level=-1.0,
    timeout=5,
    force_adjustment=1e-3,
    force_adjustment_rel_tol=1e-1,
    force_adjustment_pos_voltage_threshold=0,
    force_adjustment_neg_voltage_threshold=0,
    failure_iteration_threshold=1000,
    simulate_neighbours=True,
):
    """Method to program (alter) the conductance of a given device within a crossbar.

    Parameters
    ----------
    crossbar : memtorch.bh.crossbar.Crossbar
        Crossbar containing the device to program.
    point : tuple
        Point to program (row, column).
    conductance : float
        Conductance to program.
    rel_tol : float
        Relative tolerance between the desired conductance and the device's conductance.
    pulse_duration : float
        Duration of the programming pulse (s).
    refactory_period : float
        Duration of the refactory period (s).
    pos_voltage_level : float
        Positive voltage level (V).
    neg_voltage_level : float
        Negative voltage level (V).
    timeout : int
        Timeout (seconds) until stuck devices are unstuck.
    force_adjustment : float
        Adjustment (resistance) to unstick stuck devices.
    force_adjustment_rel_tol : float
        Relative tolerance threshold between a stuck device's conductance and high and low conductance states to force adjust.
    force_adjustment_pos_voltage_threshold : float
        Positive voltage level threshold (V) to enable force adjustment.
    force_adjustment_neg_voltage_threshold : float
        Negative voltage level threshold (V) to enable force adjustment.
    failure_iteration_threshold : int
        Failure iteration threshold.
    simulate_neighbours : bool
        Simulate neighbours (True).

    Returns
    -------
    memtorch.bh.memristor.Memristor.Memristor
        Programmed device.
    """
    assert (1 / conductance) >= crossbar.devices[
        point
    ].r_on and conductance <= crossbar.devices[
        point
    ].r_off, "Conductance to program must be between g_off and g_on."
    assert (
        len(crossbar.devices.shape) == 2 or len(crossbar.devices.shape) == 3
    ), "Invalid devices shape."
    if len(crossbar.devices.shape) == 3:
        tile, row, column = point
    else:
        row, column = point
        tile = None

    time_signal, pos_voltage_signal = gen_programming_signal(
        1,
        pulse_duration,
        refactory_period,
        pos_voltage_level,
        crossbar.devices[point].time_series_resolution,
    )
    _, neg_voltage_signal = gen_programming_signal(
        1,
        pulse_duration,
        refactory_period,
        neg_voltage_level,
        crossbar.devices[point].time_series_resolution,
    )
    timeout = time.time() + timeout
    iterations = 0
    while not math.isclose(conductance, crossbar.devices[point].g, rel_tol=rel_tol):
        if conductance < crossbar.devices[point].g:
            voltage_signal = neg_voltage_signal
        else:
            voltage_signal = pos_voltage_signal

        previous_g = crossbar.devices[point].g
        crossbar.devices[point].simulate(voltage_signal)
        if simulate_neighbours:
            for row_ in range(0, crossbar.devices.shape[-2]):
                if row_ != row:
                    if tile is not None:
                        idx = (tile, row_, column)
                    else:
                        idx = (row_, column)

                    crossbar.devices[idx].simulate(voltage_signal / 2)

            for column_ in range(0, crossbar.devices.shape[-1]):
                if column_ != column:
                    if tile is not None:
                        idx = (tile, row, column_)
                    else:
                        idx = (row, column_)

                    crossbar.devices[idx].simulate(voltage_signal / 2)

        if crossbar.devices[point].g == previous_g:
            if (
                np.amax(voltage_signal) >= force_adjustment_pos_voltage_threshold
                or np.amin(voltage_signal) <= force_adjustment_neg_voltage_threshold
            ):
                if math.isclose(
                    previous_g,
                    1 / crossbar.devices[point].r_on,
                    rel_tol=force_adjustment_rel_tol,
                ):
                    crossbar.devices[point].set_conductance(
                        crossbar.devices[point].g - force_adjustment
                    )
                elif math.isclose(
                    previous_g,
                    1 / crossbar.devices[point].r_off,
                    rel_tol=force_adjustment_rel_tol,
                ):
                    crossbar.devices[point].set_conductance(
                        crossbar.devices[point].g + force_adjustment
                    )

        iterations += 1
        if iterations >= failure_iteration_threshold and time.time() > timeout:
            warnings.warn("Failed to program device to rel_tol (%f)." % rel_tol)
            break

    return crossbar.devices


def gen_programming_signal(
    number_of_pulses,
    pulse_duration,
    refactory_period,
    voltage_level,
    time_series_resolution,
):
    """Method to generate a programming signal using a sequence of pulses.

    Parameters
    ----------
    number_of_pulses : int
        Number of pulses.
    pulse_duration : float
        Duration of the programming pulse (s).
    refactory_period : float
        Duration of the refactory period (s).
    voltage_level : float
        Voltage level (V).
    time_series_resolution : float
        Time series resolution (s).

    Returns
    -------
    tuple
        Tuple containing the generated time and voltage signals.
    """
    period = pulse_duration + refactory_period
    duration = number_of_pulses * period
    assert_tol = 1e-9
    assert (
        abs(
            pulse_duration / time_series_resolution
            - round(pulse_duration / time_series_resolution)
        )
        <= assert_tol
    ), "pulse_duration must be divisible by time_series_resolution."
    assert (
        abs(
            refactory_period / time_series_resolution
            - round(refactory_period / time_series_resolution)
        )
        <= assert_tol
    ), "refactory_period must be divisible by time_series_resolution."
    time_signal = np.arange(0, duration, step=time_series_resolution)
    period = np.zeros(round(period / time_series_resolution))
    period[0 : round(pulse_duration / time_series_resolution)] = voltage_level
    voltage_signal = np.tile(period, number_of_pulses)
    return time_signal, voltage_signal



import copy
import math
import random

import numpy as np
import pytest
import torch

import memtorch


@pytest.mark.parametrize("x", [0, 0.25, 0.5, 0.75, 1.0])
def test_window_functions(x):
    for method in dir(memtorch.bh.memristor.window):
        if "__" in method:
            break

        method = getattr(memtorch.bh.memristor.window, method)
        assert callable(method)
        assert type(method(x)) == float



import math
from abc import ABC, abstractmethod

import matplotlib.pyplot as plt
import numpy as np
import torch
from scipy import signal

import memtorch


class Memristor(ABC):
    """
    Parameters
    ----------
    r_off : float
        Off (maximum) resistance of the device (ohms).
    r_on : float
        On (minimum) resistance of the device (ohms).
    time_series_resolution : float
        Time series resolution (s).
    pos_write_threshold : float
        Positive write threshold voltage (V).
    neg_write_threshold : float
        Negative write threshold voltage (V).
    """

    def __init__(
        self,
        r_off,
        r_on,
        time_series_resolution,
        pos_write_threshold=0,
        neg_write_threshold=0,
    ):
        self.r_off = r_off
        self.r_on = r_on
        self.time_series_resolution = time_series_resolution
        self.pos_write_threshold = pos_write_threshold
        self.neg_write_threshold = neg_write_threshold
        self.g = None
        self.finite_states = None

    @abstractmethod
    def simulate(self, voltage_signal):
        """Method to determine the equivalent conductance of a memristive device when a given voltage signal is applied.

        Parameters
        ----------
        voltage_signal : torch.Tensor
            A discrete voltage signal with resolution time_series_resolution.

        Returns
        -------
        torch.Tensor
            A tensor containing the equivalent device conductance for each simulated timestep.
        """
        return

    @abstractmethod
    def set_conductance(self, conductance):
        """Method to manually set the conductance of a memristive device.

        Parameters
        ----------
        conductance : float
                Conductance to set.
        """
        return

    def get_resistance(self):
        """
        Method to determine the resistance of a memristive device.

        Returns
        -------
        float
            The devices resistance (ohms).
        """
        return 1 / self.g

    @abstractmethod
    def plot_hysteresis_loop(
        self,
        memristor,
        duration,
        voltage_signal_amplitude,
        voltage_signal_frequency,
        log_scale=False,
        return_result=False,
    ):
        """Method to plot the hysteresis loop of a given device.

        Parameters
        ----------
        memristor : memtorch.bh.memristor.Memristor.Memristor
            Memristor.
        duration : float
            Duration (s).
        voltage_signal_amplitude: float
            Voltage signal amplitude (V).
        voltage_signal_frequency : float
            Voltage signal frequency (Hz)
        log_scale : bool
            Plot the y-axis (current) using a symmetrical log scale (True).
        return_result: bool
            Voltage and current signals are returned (True).
        """
        return plot_hysteresis_loop(
            memristor,
            duration,
            voltage_signal_amplitude,
            voltage_signal_frequency,
            log_scale,
            return_result,
        )

    @abstractmethod
    def plot_bipolar_switching_behaviour(
        self,
        memristor,
        voltage_signal_amplitude,
        voltage_signal_frequency,
        log_scale=True,
        return_result=False,
    ):
        """Method to plot the DC bipolar switching behaviour of a given device.

        Parameters
        ----------
        memristor : memtorch.bh.memristor.Memristor.Memristor
            Memristor.
        voltage_signal_amplitude: float
            Voltage signal amplitude (V).
        voltage_signal_frequency : float
            Voltage signal frequency (Hz)
        log_scale : bool
            Plot the y-axis (current) using a symmetrical log scale (True).
        return_result: bool
            Voltage and current signals are returned (True).
        """
        return plot_bipolar_switching_behaviour(
            memristor,
            voltage_signal_amplitude,
            voltage_signal_frequency,
            log_scale,
            return_result,
        )


def plot_hysteresis_loop(
    memristor_model,
    duration,
    voltage_signal_amplitude,
    voltage_signal_frequency,
    log_scale=False,
    return_result=False,
):
    """Method to plot the hysteresis loop of a given device.

    Parameters
    ----------
    memristor_model : memtorch.bh.memristor.Memristor.Memristor
        Memristor model.
    duration : float
        Duration (s).
    voltage_signal_amplitude: float
        Voltage signal amplitude (V).
    voltage_signal_frequency : float
        Voltage signal frequency (Hz)
    log_scale : bool
        Plot the y-axis (current) using a symmetrical log scale (True).
    return_result: bool
        Voltage and current signals are returned (True).

    Returns
    -------
    tuple
        Voltage and current signals.
    """
    time_signal = np.arange(
        0,
        duration + memristor_model.time_series_resolution,
        step=memristor_model.time_series_resolution,
    )
    voltage_signal = voltage_signal_amplitude * np.sin(
        2 * math.pi * voltage_signal_frequency * time_signal
    )
    current_signal = memristor_model.simulate(voltage_signal, return_current=True)
    if return_result:
        return voltage_signal, current_signal
    else:
        plt.figure()
        plt.title("Hysteresis Loop")
        plt.xlabel("Voltage (V)")
        plt.plot(voltage_signal, current_signal)
        if log_scale:
            plt.ylabel("log10(Current (A))")
            plt.yscale("symlog")
        else:
            plt.ylabel("Current (A)")

        plt.show()
        return


def plot_bipolar_switching_behaviour(
    memristor_model,
    voltage_signal_amplitude,
    voltage_signal_frequency,
    log_scale=True,
    return_result=False,
):
    """Method to plot the DC bipolar switching behaviour of a given device.

    Parameters
    ----------
    memristor_model : memtorch.bh.memristor.Memristor.Memristor
        Memristor model.
    voltage_signal_amplitude: float
        Voltage signal amplitude (V).
    voltage_signal_frequency : float
        Voltage signal frequency (Hz)
    log_scale : bool
        Plot the y-axis (current) using a symmetrical log scale (True).
    return_result: bool
        Voltage and current signals are returned (True).
    """

    def gen_triangle_waveform(n_points, amplitude):
        def triangle_iterator_generator(n, amp):
            y = 0
            x = 0
            s = amplitude / (n / 4)
            while x < n_points:
                yield y
                y += s
                if abs(y) > amplitude:
                    s *= -1

                x += 1

        return np.fromiter(triangle_iterator_generator(n_points, amplitude), "d")

    time_signal = np.arange(
        0,
        (1 / voltage_signal_frequency) + memristor_model.time_series_resolution,
        step=memristor_model.time_series_resolution,
    )
    voltage_signal = gen_triangle_waveform(len(time_signal), voltage_signal_amplitude)
    current_signal = memristor_model.simulate(voltage_signal, return_current=True)
    if return_result:
        return voltage_signal, current_signal
    else:
        plt.figure()
        plt.title("Bipolar Switching Behaviour (DC)")
        plt.xlabel("Voltage (V)")
        if log_scale:
            plt.ylabel("|log10(Current (A))|")
            plt.yscale("log")
        else:
            plt.ylabel("|Current (A)|")

        plt.plot(voltage_signal, abs(current_signal))
        plt.show()
        return



import copy

import pytest
import torch

import memtorch
from memtorch.bh.crossbar.Program import naive_program
from memtorch.map.Parameter import naive_map
from memtorch.mn.Module import patch_model


@pytest.mark.parametrize("tile_shape", [None, (128, 128), (10, 20)])
def test_schemes(debug_networks, tile_shape):
    networks = debug_networks
    for scheme in memtorch.bh.Scheme:
        for network in networks:
            patched_network = patch_model(
                copy.deepcopy(network),
                memristor_model=memtorch.bh.memristor.LinearIonDrift,
                memristor_model_params={},
                module_parameters_to_patch=[type(network.layer)],
                mapping_routine=naive_map,
                transistor=True,
                programming_routine=None,
                scheme=scheme,
                tile_shape=tile_shape,
            )
            assert patched_network.layer.crossbars is not None



import numpy as np
import torch

import memtorch
from memtorch.bh.crossbar.Program import naive_program
from memtorch.bh.nonideality.endurance_retention_models.empirical_metal_oxide_RRAM import (
    model_endurance_retention,
)
from memtorch.map.Parameter import naive_map

# TEMP for if __name__ == "__main__"
from memtorch.mn.Module import patch_model

valid_endurance_models = [model_endurance_retention]


def apply_endurance_model(
    layer, x, endurance_model=model_endurance_retention, **endurance_model_kwargs
):
    """Method to apply an endurance model to devices within a memristive layer.

    Parameters
    ----------
    layer : memtorch.mn
        A memrstive layer.
    x : float
        Energy (J) / SET-RESET cycles.
    endurance_model : function
        Endurance model to use.
    endurance_model_kwargs : **kwargs
        Endurance model keyword arguments.

    Returns
    -------
    memtorch.mn
        The patched memristive layer.
    """
    assert endurance_model in valid_endurance_models, "endurance_model is not valid."
    return endurance_model(layer=layer, x=x, **endurance_model_kwargs)



import copy
import math

import matplotlib.pyplot as plt
import numpy as np
import torch

import memtorch


def apply_non_linear(
    layer,
    sweep_duration=1,
    sweep_voltage_signal_amplitude=1,
    sweep_voltage_signal_frequency=1,
    num_conductance_states=None,
    simulate=False,
):
    """Method to model non_linear I/V characteristics for devices within a memristive layer.

    Parameters
    ----------
    layer : memtorch.mn
        A memrstive layer.
    sweep_duration : float
        Voltage sweep duration (s).
    sweep_voltage_signal_amplitude : float
        Voltage sweep amplitude (V).
    sweep_voltage_signal_frequency : float
        Voltage sweep frequency (Hz).
    num_conductance_states : int, optional
        Number of finite conductance states to model. None indicates finite states are not to be modeled.
    simulate : bool, optional
        Each device is simulated during inference (True).

    Returns
    -------
    memtorch.mn
        The patched memristive layer.
    """

    def apply_non_linear_to_device(
        device,
        sweep_duration,
        sweep_voltage_signal_amplitude,
        sweep_voltage_signal_frequency,
    ):
        time_signal = np.arange(
            0,
            sweep_duration + device.time_series_resolution,
            step=device.time_series_resolution,
        )
        voltage_signal = np.cos(
            2 * math.pi * sweep_voltage_signal_frequency * time_signal
        )
        current_signal = copy.deepcopy(device).simulate(
            voltage_signal, return_current=True
        )

        def det_current(voltage):
            if np.isnan(voltage.cpu()):
                return 0

            assert (
                abs(voltage) <= sweep_voltage_signal_amplitude
            ), "voltage must be between -sweep_voltage_signal_amplitude and sweep_voltage_signal_amplitude."
            if voltage < 0:
                return (
                    -1
                    * current_signal[::-1][
                        np.searchsorted(
                            voltage_signal[::-1], -1 * voltage.cpu(), side="left"
                        )
                    ]
                )
            else:
                return current_signal[::-1][
                    np.searchsorted(voltage_signal[::-1], voltage.cpu(), side="left")
                ]

        device.det_current = det_current
        return device

    def apply_non_linear_to_crossbar(
        crossbar,
        sweep_duration,
        sweep_voltage_signal_amplitude,
        sweep_voltage_signal_frequency,
    ):
        assert (
            len(crossbar.devices.shape) == 2 or len(crossbar.devices.shape) == 3
        ), "Invalid devices shape."
        if len(crossbar.devices.shape) == 2:
            for row in range(0, crossbar.rows):
                for column in range(0, crossbar.columns):
                    crossbar.devices[row, column] = apply_non_linear_to_device(
                        crossbar.devices[row, column],
                        sweep_duration,
                        sweep_voltage_signal_amplitude,
                        sweep_voltage_signal_frequency,
                    )
        else:
            for i in range(0, crossbar.devices.shape[0]):
                for j in range(crossbar.devices.shape[1]):
                    for k in range(crossbar.devices.shape[2]):
                        crossbar.devices[i, j, k] = apply_non_linear_to_device(
                            crossbar.devices[i, j, k],
                            sweep_duration,
                            sweep_voltage_signal_amplitude,
                            sweep_voltage_signal_frequency,
                        )

        return crossbar

    layer.non_linear = True
    if simulate:
        layer.simulate = True
    else:
        if num_conductance_states is None:
            for i in range(len(layer.crossbars)):
                layer.crossbars[i] = apply_non_linear_to_crossbar(
                    layer.crossbars[i],
                    sweep_duration,
                    sweep_voltage_signal_amplitude,
                    sweep_voltage_signal_frequency,
                )
        else:
            raise ("To be implemented.")

    return layer



import math
from enum import Enum, auto, unique

import numpy as np
import torch

import memtorch
import memtorch.mn
from memtorch.bh.nonideality.DeviceFaults import apply_device_faults
from memtorch.bh.nonideality.Endurance import apply_endurance_model
from memtorch.bh.nonideality.FiniteConductanceStates import (
    apply_finite_conductance_states,
)
from memtorch.bh.nonideality.NonLinear import apply_non_linear
from memtorch.bh.nonideality.Retention import apply_retention_model
from memtorch.mn.Module import supported_module_parameters


@unique
class NonIdeality(Enum):
    """NonIdeality enumeration."""

    FiniteConductanceStates = auto()
    DeviceFaults = auto()
    NonLinear = auto()
    Endurance = auto()
    Retention = auto()


def apply_nonidealities(model, non_idealities, **kwargs):
    """Method to apply non-idealities to a torch.nn.Module instance with memristive layers.

    Parameters
    ----------
    model : torch.nn.Module
        torch.nn.Module instance.
    nonidealities : memtorch.bh.nonideality.NonIdeality.NonIdeality, tuple
        Non-linearitites to model.

    Returns
    -------
    torch.nn.Module
        Patched instance.
    """

    def apply_patched_module(model, patched_module, name, m):
        model._modules[name] = patched_module
        return model

    for _, (name, m) in enumerate(list(model.named_modules())):
        if type(m) in supported_module_parameters.values():
            for non_ideality in non_idealities:
                if non_ideality == NonIdeality.FiniteConductanceStates:
                    required(
                        kwargs,
                        ["conductance_states"],
                        "memtorch.bh.nonideality.NonIdeality.FiniteConductanceStates",
                    )
                    model = apply_patched_module(
                        model,
                        apply_finite_conductance_states(
                            m, kwargs["conductance_states"]
                        ),
                        name,
                        m,
                    )
                elif non_ideality == NonIdeality.DeviceFaults:
                    required(
                        kwargs,
                        ["lrs_proportion", "hrs_proportion", "electroform_proportion"],
                        "memtorch.bh.nonideality.NonIdeality.DeviceFaults",
                    )
                    model = apply_patched_module(
                        model,
                        apply_device_faults(
                            m,
                            kwargs["lrs_proportion"],
                            kwargs["hrs_proportion"],
                            kwargs["electroform_proportion"],
                        ),
                        name,
                        m,
                    )
                elif non_ideality == NonIdeality.NonLinear:
                    if "simulate" in kwargs:
                        if kwargs["simulate"] == True:
                            model = apply_patched_module(
                                model, apply_non_linear(m, simulate=True), name, m
                            )
                        else:
                            required(
                                kwargs,
                                [
                                    "sweep_duration",
                                    "sweep_voltage_signal_amplitude",
                                    "sweep_voltage_signal_frequency",
                                ],
                                "memtorch.bh.nonideality.NonIdeality.NonLinear",
                            )
                            model = apply_patched_module(
                                model,
                                apply_non_linear(
                                    m,
                                    kwargs["sweep_duration"],
                                    kwargs["sweep_voltage_signal_amplitude"],
                                    kwargs["sweep_voltage_signal_frequency"],
                                ),
                                name,
                                m,
                            )
                    else:
                        required(
                            kwargs,
                            [
                                "sweep_duration",
                                "sweep_voltage_signal_amplitude",
                                "sweep_voltage_signal_frequency",
                            ],
                            "memtorch.bh.nonideality.NonIdeality.NonLinear",
                        )
                        model = apply_patched_module(
                            model,
                            apply_non_linear(
                                m,
                                kwargs["sweep_duration"],
                                kwargs["sweep_voltage_signal_amplitude"],
                                kwargs["sweep_voltage_signal_frequency"],
                            ),
                            name,
                            m,
                        )
                elif non_ideality == NonIdeality.Endurance:
                    required(
                        kwargs,
                        ["x", "endurance_model", "endurance_model_kwargs"],
                        "memtorch.bh.nonideality.Endurance",
                    )
                    model = apply_patched_module(
                        model,
                        apply_endurance_model(
                            layer=m,
                            x=kwargs["x"],
                            endurance_model=kwargs["endurance_model"],
                            **kwargs["endurance_model_kwargs"]
                        ),
                        name,
                        m,
                    )
                elif non_ideality == NonIdeality.Retention:
                    required(
                        kwargs,
                        ["time", "retention_model", "retention_model_kwargs"],
                        "memtorch.bh.nonideality.Retention",
                    )
                    model = apply_patched_module(
                        model,
                        apply_retention_model(
                            layer=m,
                            time=kwargs["time"],
                            retention_model=kwargs["retention_model"],
                            **kwargs["retention_model_kwargs"]
                        ),
                        name,
                        m,
                    )

    return model


def required(kwargs, arguments, call):
    """Method to check is required arguments in **kwargs are present.

    Parameters
    ----------
    kwargs : **kwargs
        Keyword-arguments.
    arguments : list of str
        Arguments which are required to be present.
    call : str
        Function to call.
    """
    for argument in arguments:
        assert kwargs[argument] is not None, "%s is required when calling %s" % (
            argument,
            call,
        )



"""
Conductance drift model as described in:
I. Boybat et al., "Impact of conductance drift on multi-PCM synaptic architectures", 2018 Non-Volatile Memory Technology Symposium (NVMTS), 2018.
"""
import numpy as np
import torch


def model_conductance_drift(layer, x, initial_time=1e-12, drift_coefficient=0.1):
    """
    Parameters
    ----------
    layer : memtorch.mn
        A memrstive layer.
    x : float
        Retention time (s). Denoted using x for sake of consistency with other models.
    initial_time : float
        Initial time that corresponds with initial conductance values.
    drift_coefficient : float
        Drift coefficient. For PCM devices, drift_coefficient typically has a value of 0.1 for the amorphous phase.

    Returns
    -------
    memtorch.mn
        The patched memristive layer.
    """
    time = x
    if initial_time == 0.0:
        initial_time = 1e-12

    assert (
        time >= 0 and initial_time >= 0 and time > initial_time
    ), "time and/or initial_time are/is invalid."
    assert (
        drift_coefficient >= 0 and drift_coefficient <= 1
    ), "drift_coefficient must be >=0 and <= 1."
    for i in range(len(layer.crossbars)):
        initial_conductance = layer.crossbars[i].conductance_matrix
        initial_conductance = initial_conductance * (
            (time / initial_time) ** (-drift_coefficient)
        )
        layer.crossbars[i].conductance_matrix = initial_conductance

    return layer



import math

import numpy as np
import torch

import memtorch
from memtorch.utils import clip, convert_range

from .Memristor import Memristor as Memristor


class LinearIonDrift(Memristor):
    """Linear Ion behvaioural drift model.

    Parameters
    ----------
    time_series_resolution : float
        Time series resolution (s).
    u_v : float
        Dopant drift mobility of the device material.
    d : float
        Device length (m).
    r_on : float
        On (minimum) resistance of the device (ohms).
    r_off : float
        Off (maximum) resistance of the device (ohms).
    pos_write_threshold: float
        Positive write threshold voltage (V).
    neg_write_threshold : float
        Negative write threshold voltage (V).
    p : int
        Joglekar window p constant.
    """

    def __init__(
        self,
        time_series_resolution=1e-4,
        u_v=1e-14,
        d=10e-9,
        r_on=100,
        r_off=16e3,
        pos_write_threshold=0.55,
        neg_write_threshold=-0.55,
        p=1,
        **kwargs
    ):

        args = memtorch.bh.unpack_parameters(locals())
        super(LinearIonDrift, self).__init__(
            args.r_off,
            args.r_on,
            args.time_series_resolution,
            args.pos_write_threshold,
            args.neg_write_threshold,
        )
        self.u_v = args.u_v
        self.d = args.d
        self.r_i = args.r_on
        self.p = args.p
        self.g = 1 / self.r_i
        self.x = convert_range(self.r_i, self.r_on, self.r_off, 0, 1)

    def simulate(self, voltage_signal, return_current=False):
        len_voltage_signal = 1
        try:
            len_voltage_signal = len(voltage_signal)
        except:
            voltage_signal = [voltage_signal]

        if return_current:
            current = np.zeros(len_voltage_signal)

        np.seterr(all="raise")
        for t in range(0, len_voltage_signal):
            current_ = self.current(voltage_signal[t])
            if (
                voltage_signal[t] >= self.pos_write_threshold
                or voltage_signal[t] <= self.neg_write_threshold
            ):
                self.x = self.x + self.dxdt(current_) * self.time_series_resolution
                self.x = max(min(1.0, self.x), 0.0)

            try:
                self.g = 1 / ((self.r_on * self.x) + (self.r_off * (1 - self.x)))
            except:
                self.g = 0

            if self.g > (1 / self.r_on):
                self.g = 1 / self.r_on
            elif self.g < (1 / self.r_off):
                self.g = 1 / self.r_off

            if return_current:
                current[t] = current_

        if return_current:
            return current

    def set_conductance(self, conductance):
        conductance = clip(conductance, 1 / self.r_off, 1 / self.r_on)
        self.x = convert_range(1 / conductance, self.r_on, self.r_off, 0, 1)
        self.g = conductance

    def current(self, voltage):
        """Method to determine the current of the model given an applied voltage.

        Parameters
        ----------
        voltage : float
            The current applied voltage (V).

        Returns
        -------
        float
            The observed current (A).
        """
        return voltage / (((self.r_on * self.x) + (self.r_off * (1 - self.x))))

    def dxdt(self, current):
        """Method to determine the derivative of the state variable, dx/dt.

        Parameters
        ----------
        current : float
            The observed current (A).

        Returns
        -------
        float
            The derivative of the state variable, dx/dt.
        """
        return (
            self.u_v
            * (self.r_on / (self.d**2))
            * current
            * memtorch.bh.memristor.window.Jogelkar(self.x, self.p)
        )

    def plot_hysteresis_loop(
        self,
        duration=4,
        voltage_signal_amplitude=5,
        voltage_signal_frequency=2.5,
        return_result=False,
    ):
        return super().plot_hysteresis_loop(
            self,
            duration=duration,
            voltage_signal_amplitude=voltage_signal_amplitude,
            voltage_signal_frequency=voltage_signal_frequency,
            return_result=return_result,
        )

    def plot_bipolar_switching_behaviour(
        self,
        voltage_signal_amplitude=5,
        voltage_signal_frequency=2.5,
        log_scale=True,
        return_result=False,
    ):
        return super().plot_bipolar_switching_behaviour(
            self,
            voltage_signal_amplitude=voltage_signal_amplitude,
            voltage_signal_frequency=voltage_signal_frequency,
            log_scale=log_scale,
            return_result=return_result,
        )



import copy
import inspect
import math

import torch

import memtorch


def StochasticParameter(
    distribution=torch.distributions.normal.Normal,
    min=0,
    max=float("Inf"),
    function=True,
    **kwargs
):
    """Method to model a stochastic parameter.

    Parameters
    ----------
    distribution : torch.distributions
        torch distribution.
    min : float
        Minimum value to sample.
    max: float
        Maximum value to sample.
    function : bool
        A sampled value is returned (False). A function to return a sampled value or mean is returned (True).

    Returns
    -------
    float or function
        A sampled value of the stochatic parameter, or a sample-value generator.
    """
    assert issubclass(
        distribution, torch.distributions.distribution.Distribution
    ), "Distribution is not in torch.distributions."
    for arg in inspect.signature(distribution).parameters.values():
        if arg.name not in kwargs and arg.name != "validate_args":
            raise Exception("Argument %s is required for %s" % (arg.name, distribution))

    m = distribution(**kwargs)

    def f(return_mean=False):
        """Method to return a sampled value or the mean of the stochatic parameters.

        Parameters
        ----------
        return_mean : bool
            Return the mean value of the stochatic parameter (True). Return a sampled value of the stochatic parameter (False).

        Returns
        -------
        float
            The mean value, or a sampled value of the stochatic parameter.
        """
        if return_mean:
            return m.mean
        else:
            return m.sample().clamp(min, max).item()

    if function:
        return f
    else:
        return f()


def unpack_parameters(local_args, r_rel_tol=None, r_abs_tol=None, resample_threshold=5):
    """Method to sample from stochastic sample-value generators

    Parameters
    ----------
    local_args : locals()
        Local arguments with stochastic sample-value generators from which to sample from.
    r_rel_tol : float
        Relative threshold tolerance.
    r_abs_tol : float
        Absolute threshold tolerance.
    resample_threshold : int
        Number of times to resample r_off and r_on when their proximity is within the threshold tolerance before raising an exception.

    Returns
    -------
    **
        locals() with sampled stochastic parameters.
    """
    assert (
        r_rel_tol is None or r_abs_tol is None
    ), "r_rel_tol or r_abs_tol must be None."
    assert (
        type(resample_threshold) == int and resample_threshold >= 0
    ), "resample_threshold must be of type int and >= 0."
    if "reference" in local_args:
        return_mean = True
    else:
        return_mean = False

    local_args_copy = copy.deepcopy(local_args)
    for arg in local_args:
        if callable(local_args[arg]) and "__" not in str(arg):
            local_args[arg] = local_args[arg](return_mean=return_mean)

    args = Dict2Obj(local_args)
    if hasattr(args, "r_off") and hasattr(args, "r_on"):
        resample_idx = 0
        r_off_generator = local_args_copy["r_off"]
        r_on_generator = local_args_copy["r_on"]
        while True:
            if r_abs_tol is None and r_rel_tol is not None:
                if not math.isclose(args.r_off, args.r_on, rel_tol=r_rel_tol):
                    break
            elif r_rel_tol is None and r_abs_tol is not None:
                if not math.isclose(args.r_off, args.r_on, abs_tol=r_abs_tol):
                    break
            else:
                if not math.isclose(args.r_off, args.r_on):
                    break

            if callable(r_off_generator) and callable(r_on_generator):
                args.r_off = copy.deepcopy(r_off_generator)(return_mean=return_mean)
                args.r_on = copy.deepcopy(r_on_generator)(return_mean=return_mean)
            else:
                raise Exception(
                    "Resample threshold exceeded (deterministic values used)."
                )

            resample_idx += 1
            if resample_idx > resample_threshold:
                raise Exception("Resample threshold exceeded.")

    return args


class Dict2Obj(object):
    """Class used to instantiate a object given a dictionary."""

    def __init__(self, dictionary):
        for key in dictionary:
            if key == "__class__":
                continue

            setattr(self, key, dictionary[key])



import pytest
import torch

import memtorch
from memtorch.utils import LoadCIFAR10, LoadMNIST


@pytest.mark.parametrize("dataloader", [LoadMNIST, LoadCIFAR10])
def test_dataloader(dataloader):
    train_loader, validation_loader, test_loader = dataloader(validation=True)
    assert train_loader is not None
    assert validation_loader is not None
    assert test_loader is not None
    train_loader, validation_loader, test_loader = dataloader(validation=False)
    assert train_loader is not None
    assert validation_loader is None
    assert test_loader is not None



import math

import numpy as np
import torch

import memtorch


def Biolek(voltage, x=1.0, p=1):
    """Biolek window function.

    Parameters
    ----------
    voltage : float
        Input voltage (V).
    x : float
        State variable.
    p : int
        p constant.
    """

    def step(x):
        return 1 * (x > 0)

    return 1 - (1 - x - step(voltage)) ** (2 * p)



import copy
import math

import numpy as np
import pytest
import torch
from numpy.lib import source

import memtorch
from memtorch.bh.crossbar.Program import naive_program
from memtorch.map.Parameter import naive_map
from memtorch.mn.Module import patch_model


@pytest.mark.parametrize("tile_shape", [None, (128, 128), (10, 20)])
@pytest.mark.parametrize("quant_method", memtorch.bh.Quantize.quant_methods + [None])
@pytest.mark.parametrize("source_resistance", [None, 5])
@pytest.mark.parametrize("line_resistance", [None, 5])
@pytest.mark.parametrize("use_bindings", [True, False])
def test_networks(
    debug_networks,
    tile_shape,
    quant_method,
    source_resistance,
    line_resistance,
    use_bindings,
):
    networks = debug_networks
    if quant_method is not None:
        ADC_resolution = 8
    else:
        ADC_resolution = None

    for network in networks:
        patched_network = patch_model(
            copy.deepcopy(network),
            memristor_model=memtorch.bh.memristor.LinearIonDrift,
            memristor_model_params={},
            module_parameters_to_patch=[type(network.layer)],
            mapping_routine=naive_map,
            transistor=True,
            programming_routine=None,
            scheme=memtorch.bh.Scheme.SingleColumn,
            tile_shape=tile_shape,
            max_input_voltage=1.0,
            ADC_resolution=ADC_resolution,
            quant_method=quant_method,
            source_resistance=source_resistance,
            line_resistance=line_resistance,
            use_bindings=use_bindings,
        )
        patched_network.tune_()
        patched_network.disable_legacy()



import numpy as np
import torch
import torch.functional as F
import torch.nn as nn

import memtorch
from memtorch.utils import convert_range


def naive_scale(module, input, force_scale=False):
    """Naive method to encode input values as bit-line voltages.

    Parameters
    ----------
    module : torch.nn.Module
        Memristive layer to tune.
    input : torch.tensor
        Input tensor to encode.
    force_scale : bool, optional
        Used to determine if inputs are scaled (True) or not (False) if they no not exceed max_input_voltage.

    Returns
    -------
    torch.Tensor
        Encoded voltages.
    """
    if module.max_input_voltage is not None:
        assert (
            type(module.max_input_voltage) == int
            or type(module.max_input_voltage) == float
        ) and module.max_input_voltage > 0, (
            "The maximum input voltage (max_input_voltage) must be >0."
        )
        input_range = torch.amax(torch.abs(input))
        if not force_scale and input_range <= module.max_input_voltage:
            return input
        else:
            return convert_range(
                input,
                -input_range,
                input_range,
                -module.max_input_voltage,
                module.max_input_voltage,
            )

    return input



import itertools
import multiprocessing as mp

import torch
import torch.functional as F
from torch.nn import modules
from torch.nn.modules import module

import memtorch
from memtorch.map.Input import naive_scale
from memtorch.map.Parameter import naive_map

from .Conv1d import Conv1d
from .Conv2d import Conv2d
from .Conv3d import Conv3d
from .Linear import Linear

supported_module_parameters = {
    "<class 'torch.nn.modules.linear.Linear'>": Linear,
    "<class 'torch.nn.modules.conv.Conv1d'>": Conv1d,
    "<class 'torch.nn.modules.conv.Conv2d'>": Conv2d,
    "<class 'torch.nn.modules.conv.Conv3d'>": Conv3d,
}


def patch_model(
    model,
    memristor_model,
    memristor_model_params,
    module_parameters_to_patch={},
    mapping_routine=naive_map,
    transistor=True,
    programming_routine=None,
    programming_routine_params={"rel_tol": 0.1},
    p_l=None,
    scheme=memtorch.bh.Scheme.DoubleColumn,
    tile_shape=None,
    max_input_voltage=None,
    scaling_routine=naive_scale,
    scaling_routine_params={},
    source_resistance=None,
    line_resistance=None,
    ADC_resolution=None,
    ADC_overflow_rate=0.0,
    quant_method=None,
    use_bindings=True,
    random_crossbar_init=False,
    verbose=True,
    **kwargs
):
    """Method to convert a torch.nn model to a memristive model.

    Parameters
    ----------
    model : torch.nn.Module
        torch.nn.Module to patch.
    memristor_model : memtorch.bh.memristor.Memristor.Memristor
        Memristor model.
    memristor_model_params : **kwargs
        Memristor model keyword arguments.
    module_parameters_to_patch : module_paramter_patches
        Model parameters to patch.
    mapping_routine : function
        Mapping routine to use.
    transistor : bool
        Used to determine if a 1T1R (True) or 1R arrangement (False) is simulated.
    programming_routine : function
        Programming routine to use.
    programming_routine_params : **kwargs
        Programming routine keyword arguments.
    p_l: float
        If not None, the proportion of weights to retain.
    scheme : memtorch.bh.Scheme
        Weight representation scheme.
    tile_shape : (int, int)
        Tile shape to use to store weights. If None, modular tiles are not used.
    max_input_voltage : float
        Maximum input voltage used to encode inputs. If None, inputs are unbounded.
    scaling_routine : function
        Scaling routine to use in order to scale batch inputs.
    scaling_routine_params : **kwargs
        Scaling routine keyword arguments.
    source_resistance : float
        The resistance between word/bit line voltage sources and crossbar(s).
    line_resistance : float
        The interconnect line resistance between adjacent cells.
    ADC_resolution : int
        ADC resolution (bit width). If None, quantization noise is not accounted for.
    ADC_overflow_rate : float
        Overflow rate threshold for linear quanitzation (if ADC_resolution is not None).
    quant_method:
        Quantization method. Must be in ['linear', 'log', 'log_minmax', 'minmax', 'tanh'], or None.
    use_bindings : bool
        Used to determine if C++/CUDA bindings are used (True) or not (False).
    random_crossbar_init : bool
        Determines if the crossbar is to be initialized at random values in between Ron and Roff
    verbose : bool
        Used to determine if verbose output is enabled (True) or disabled (False).

    Returns
    -------
    torch.nn.Module
        Patched torch.nn.Module.
    """

    def patch_module(target_attr):
        parameter_type = str(type(target_attr))
        patch = supported_module_parameters.get(parameter_type)
        return patch(
            target_attr,
            memristor_model=memristor_model,
            memristor_model_params=memristor_model_params,
            mapping_routine=mapping_routine,
            transistor=transistor,
            programming_routine=programming_routine,
            programming_routine_params=programming_routine_params,
            p_l=p_l,
            scheme=scheme,
            tile_shape=tile_shape,
            max_input_voltage=max_input_voltage,
            scaling_routine=scaling_routine,
            scaling_routine_params=scaling_routine_params,
            source_resistance=source_resistance,
            line_resistance=line_resistance,
            ADC_resolution=ADC_resolution,
            ADC_overflow_rate=ADC_overflow_rate,
            quant_method=quant_method,
            use_bindings=use_bindings,
            random_crossbar_init=random_crossbar_init,
            verbose=verbose,
            **kwargs
        )

    def patch_modules(module, name=""):
        for attr_str in dir(module):
            target_attr = getattr(module, attr_str)
            if any(
                isinstance(target_attr, module_parameter)
                and not hasattr(target_attr, "transistor")
                for module_parameter in module_parameters_to_patch
            ):
                new_bn = patch_module(target_attr)
                setattr(module, attr_str, new_bn)

        if isinstance(module, torch.nn.Module):
            if type(module) == torch.nn.modules.container.Sequential:
                for idx, (name, child) in enumerate(module.named_children()):
                    if any(
                        isinstance(child, module_parameter)
                        and not hasattr(child, "transistor")
                        for module_parameter in module_parameters_to_patch
                    ):
                        target_attr = module[idx]
                        new_bn = patch_module(target_attr)
                        module[idx] = new_bn
                    else:
                        patch_modules(child, name)
            else:
                for name, child in module.named_children():
                    patch_modules(child, name)
        else:
            for child in module:
                patch_modules(child, name)

    patch_modules(model)

    def tune_(self, tune_kwargs=None):
        """Method to tune a memristive layer.
        Parameters
        ----------
        tune_kwargs : dict
            Dictionary of **kwargs for different layer types for .tune().
        """
        for _, (name, m) in enumerate(list(self.named_modules())):
            if hasattr(m, "tune"):
                if tune_kwargs is not None:
                    module_type = str(type(m))
                    if module_type in tune_kwargs:
                        m.tune(**tune_kwargs[module_type])
                    else:
                        m.tune()
                else:
                    m.tune()

    def forward_legacy(self, enable_forward_legacy):
        """Method to enable or disable forward legacy operation.

        Parameters
        ----------
        enable_forward_legacy : bool
            Enable or disable forward legacy operation.
        """
        for i, (name, m) in enumerate(list(self.named_modules())):
            if type(m) in supported_module_parameters.values():
                m.forward_legacy_enabled = enable_forward_legacy

    def disable_legacy(self):
        """Method to delete all legacy parameters to reduce memory usage. When this method is called forward_legacy is disabled."""
        for i, (name, m) in enumerate(list(self.named_modules())):
            if type(m) in supported_module_parameters.values():
                delattr(m, "weight")
                m.weight = None

        if "cpu" not in memtorch.__version__:
            torch.cuda.empty_cache()

        self.forward_legacy(False)
        delattr(self, "forward_legacy")

    def set_cuda_malloc_heap_size(self, cuda_malloc_heap_size):
        """Method to set the CUDA malloc heap size."""
        for i, (name, m) in enumerate(list(self.named_modules())):
            if type(m) in supported_module_parameters.values():
                m.cuda_malloc_heap_size = cuda_malloc_heap_size

    model.forward_legacy = forward_legacy.__get__(model)
    model.tune_ = tune_.__get__(model)
    model.forward_legacy(False)
    model.disable_legacy = disable_legacy.__get__(model)
    model.set_cuda_malloc_heap_size = set_cuda_malloc_heap_size.__get__(model)
    return model



import math

import numpy as np
import torch

import memtorch
from memtorch.utils import clip, convert_range

from .Memristor import Memristor as Memristor


class Stanford_PKU(Memristor):
    """Stanford PKU memristor model (https://nano.stanford.edu/stanford-rram-model).

    Parameters
    ----------
    time_series_resolution : float
        Time series resolution (s).
    r_off : float
        Off (maximum) resistance of the device (ohms).
    r_on : float
        On (minimum) resistance of the device (ohms).
    gap_init : float
        Initial gap distance (m).
    g_0 : float
        g_0 model parameter.
    V_0 : float
        V_0 model parameter.
    I_0 : float
        I_0 model parameter.
    read_voltage : float
        Read voltage (V) to determine the device's conductance.
    T_init : float
        Initial room tempurature.
    R_th : float
        Thermal resistance.
    gamma_init : float
        gamma_init model parameter.
    beta : float
        beta model parameter.
    t_ox : float
        Oxide thickness (m).
    F_min : float
        Minimum field requirement to enhance gap formation.
    vel_0 : float
        vel_0 model parameter.
    E_a : float
        Activation energy.
    a_0 : float
        Atom spacing.
    delta_g_init : float
        Initial delta_g value.
    model_switch : int
        Switch to select standard model (0) or dynamic model (1).
    T_crit : float
        Threshold temperature (K) for significant random variations.
    T_smth : float
        Activation energy for vacancy generation.
    """

    def __init__(
        self,
        time_series_resolution=1e-4,
        r_off=218586,
        r_on=542,
        gap_init=2e-10,
        g_0=0.25e-9,
        V_0=0.25,
        I_0=1000e-6,
        read_voltage=0.1,
        T_init=298,
        R_th=2.1e3,
        gamma_init=16,
        beta=0.8,
        t_ox=12e-9,
        F_min=1.4e9,
        vel_0=10,
        E_a=0.6,
        a_0=0.25e-9,
        delta_g_init=0.02,
        model_switch=0,
        T_crit=450,
        T_smth=500,
        **kwargs
    ):

        args = memtorch.bh.unpack_parameters(locals())
        super(Stanford_PKU, self).__init__(
            args.r_off, args.r_on, args.time_series_resolution, 0, 0
        )
        self.gap_init = args.gap_init
        self.g_0 = args.g_0
        self.V_0 = args.V_0
        self.I_0 = args.I_0
        self.read_voltage = args.read_voltage
        self.T_init = args.T_init
        self.R_th = args.R_th
        self.gamma_init = args.gamma_init
        self.beta = args.beta
        self.t_ox = args.t_ox
        self.F_min = args.F_min
        self.vel_0 = args.vel_0
        self.E_a = args.E_a
        self.a_0 = args.a_0
        self.delta_g_init = args.delta_g_init
        self.model_switch = args.model_switch
        self.T_crit = args.T_crit
        self.T_smth = args.T_smth
        gap_min = self.g_0 * np.log(
            self.I_0
            * np.sinh(self.read_voltage / self.V_0)
            / (self.read_voltage / self.r_on)
        )
        gap_max = self.g_0 * np.log(
            self.I_0
            * np.sinh(self.read_voltage / self.V_0)
            / (self.read_voltage / self.r_off)
        )
        assert (
            gap_min > 0 and gap_max > 0 and gap_max > gap_min
        ), "Invalid gap length bounds (min: %f, max: %f) encountered." % (
            gap_min,
            gap_max,
        )
        self.gap_min = gap_min
        self.gap_max = gap_max
        self.gap = max(min(gap_init, gap_max), gap_min)
        self.g = self.current(self.read_voltage) / self.read_voltage

    def current(self, voltage):
        """Method to determine the current of the model given an applied voltage.

        Parameters
        ----------
        voltage : float
            The current applied voltage (V).

        Returns
        -------
        float
            The observed current (A).
        """
        return self.I_0 * np.exp(-self.gap / self.g_0) * np.sinh(voltage / self.V_0)

    def T_current(self, voltage, current):
        """Method to determine the thermal current of the model given an applied voltage and current.

        Parameters
        ----------
        voltage : float
            The current applied voltage (V).
        current : float
            The current applied current (A).

        Returns
        -------
        float
            The observed thermal current (A).
        """
        return self.T_init + abs(voltage * current * self.R_th)

    def dg_dt(self, voltage, current):
        """Method to determine the derivative of the gap length.

        Parameters
        ----------
        voltage : float
            The current applied voltage (V).
        current : float
            The current applied current (A).

        Returns
        -------
        float
            The derivative of the gap length.
        """
        q = 1.6e-19
        k_b = 1.3806503e-23
        gamma = self.gamma_init - self.beta * np.power(self.gap / 1e-9, 3)
        if gamma * abs(voltage) / self.t_ox < self.F_min:
            gamma = 0

        delta_g = self.delta_g_init * self.model_switch
        T_current_eval = self.T_current(voltage, current)
        return -self.vel_0 * np.exp(-q * self.E_a / k_b / T_current_eval) * np.sinh(
            gamma * self.a_0 / self.t_ox * q * voltage / k_b / T_current_eval
        ) + np.random.normal(loc=0, scale=1) * delta_g / (
            1 + np.exp((self.T_crit - T_current_eval) / self.T_smth)
        )

    def simulate(self, voltage_signal, return_current=False):
        len_voltage_signal = 1
        try:
            len_voltage_signal = len(voltage_signal)
        except:
            voltage_signal = [voltage_signal]

        if return_current:
            current = np.zeros(len_voltage_signal)

        np.seterr(all="raise")
        current_ = 0
        for t in range(0, len_voltage_signal):
            self.gap = (
                self.gap
                + self.dg_dt(voltage_signal[t], current_) * self.time_series_resolution
            )
            self.gap = max(min(self.gap, self.gap_max), self.gap_min)
            if voltage_signal[t] != 0:
                self.g = current_ / voltage_signal[t]

            current_ = self.current(voltage_signal[t])
            if return_current:
                current[t] = current_

        if return_current:
            return current

    def set_conductance(self, conductance):
        conductance = clip(conductance, 1 / self.r_off, 1 / self.r_on)
        gap = self.g_0 * np.log(
            self.I_0
            * np.sinh(self.read_voltage / self.V_0)
            / (self.read_voltage / (1 / conductance))
        )
        self.gap = max(min(gap, self.gap_max), self.gap_min)

    def plot_hysteresis_loop(
        self,
        duration=0.5,
        voltage_signal_amplitude=1.5,
        voltage_signal_frequency=10,
        log_scale=False,
        return_result=False,
    ):
        return super().plot_hysteresis_loop(
            self,
            duration=duration,
            voltage_signal_amplitude=voltage_signal_amplitude,
            voltage_signal_frequency=voltage_signal_frequency,
            log_scale=log_scale,
            return_result=return_result,
        )

    def plot_bipolar_switching_behaviour(
        self,
        voltage_signal_amplitude=1.5,
        voltage_signal_frequency=0.05,
        log_scale=True,
        return_result=False,
    ):
        return super().plot_bipolar_switching_behaviour(
            self,
            voltage_signal_amplitude=voltage_signal_amplitude,
            voltage_signal_frequency=voltage_signal_frequency,
            log_scale=log_scale,
            return_result=return_result,
        )



import math

import numpy as np
import torch

import memtorch
from memtorch.utils import clip, convert_range

from .Memristor import Memristor as Memristor


class VTEAM(Memristor):
    """VTEAM memristor model (https://asic2.group/tools/memristor-models/).

    Parameters
    ----------
    time_series_resolution : float
        Time series resolution (s).
    r_off : float
        Off (maximum) resistance of the device (ohms).
    r_on : float
        On (minimum) resistance of the device (ohms).
    d : float
        Device length (m).
    k_on : float
        k_on model parameter.
    k_off: float
        k_off model parameter.
    alpha_on : float
        alpha_on model parameter.
    alpha_off : float
        alpha_off model parameter.
    v_on : float
        Positive write threshold voltage (V).
    v_off : float
        Negative write threshold voltage (V).
    x_on : float
        x_on model parameter.
    x_off : float
        x_off model parameter.
    """

    def __init__(
        self,
        time_series_resolution=1e-10,
        r_off=1000,
        r_on=50,
        d=3e-9,
        k_on=-10,
        k_off=5e-4,
        alpha_on=3,
        alpha_off=1,
        v_on=-0.2,
        v_off=0.02,
        x_on=0,
        x_off=3e-9,
        **kwargs
    ):

        args = memtorch.bh.unpack_parameters(locals())
        super(VTEAM, self).__init__(
            args.r_off, args.r_on, args.time_series_resolution, args.v_off, args.v_on
        )
        self.d = args.d
        self.k_on = args.k_on
        self.k_off = args.k_off
        self.alpha_on = args.alpha_on
        self.alpha_off = args.alpha_off
        self.v_on = args.v_on
        self.v_off = args.v_off
        self.x_on = args.x_on
        self.x_off = args.x_off
        self.g = 1 / self.r_on
        self.x = self.x_on
        self.lamda = np.log(self.r_off / self.r_on)

    def dxdt(self, voltage):
        """Method to determine the derivative of the state variable.

        Parameters
        ----------
        voltage : float
            The current applied voltage (V).

        Returns
        -------
        float
            The derivative of the state variable.
        """
        if voltage >= self.v_off:
            return self.k_off * (((voltage / self.v_off) - 1) ** self.alpha_off)
        elif voltage <= self.v_on:
            return self.k_on * (((voltage / self.v_on) - 1) ** self.alpha_on)
        else:
            return 0

    def current(self, voltage):
        """Method to determine the current of the model given an applied voltage.

        Parameters
        ----------
        voltage : float
            The current applied voltage (V).

        Returns
        -------
        float
            The observed current (A).
        """
        return voltage / (
            self.r_off * self.x / self.d + self.r_on * (1 - self.x / self.d)
        )

    def simulate(self, voltage_signal, return_current=False):
        len_voltage_signal = 1
        try:
            len_voltage_signal = len(voltage_signal)
        except:
            voltage_signal = [voltage_signal]

        if return_current:
            current = np.zeros(len_voltage_signal)

        np.seterr(all="raise")
        for t in range(0, len_voltage_signal):
            self.x = self.x + (
                self.dxdt(voltage_signal[t]) * self.time_series_resolution
            )
            if self.x >= self.d or self.x <= 0:
                if self.x >= self.d:
                    self.x = self.d
                else:
                    self.x = 0

            current_ = self.current(voltage_signal[t])
            if voltage_signal[t] != 0:
                self.g = current_ / voltage_signal[t]

            if return_current:
                current[t] = current_

        if return_current:
            return current

    def set_conductance(self, conductance):
        conductance = clip(conductance, 1 / self.r_off, 1 / self.r_on)
        self.x = self.d * ((1 / conductance) - self.r_on) / (self.r_off - self.r_on)
        self.g = conductance

    def plot_hysteresis_loop(
        self,
        duration=200e-9,
        voltage_signal_amplitude=1,
        voltage_signal_frequency=50e6,
        return_result=False,
    ):
        return super(VTEAM, self).plot_hysteresis_loop(
            self,
            duration=duration,
            voltage_signal_amplitude=voltage_signal_amplitude,
            voltage_signal_frequency=voltage_signal_frequency,
            return_result=return_result,
        )

    def plot_bipolar_switching_behaviour(
        self,
        voltage_signal_amplitude=1.5,
        voltage_signal_frequency=50e6,
        log_scale=True,
        return_result=False,
    ):
        return super().plot_bipolar_switching_behaviour(
            self,
            voltage_signal_amplitude=voltage_signal_amplitude,
            voltage_signal_frequency=voltage_signal_frequency,
            log_scale=log_scale,
            return_result=return_result,
        )



import math

import numpy as np

import memtorch
from memtorch.utils import clip
from .Memristor import Memristor as Memristor


class Data_Driven(Memristor):
    """A Data-Driven Verilog-A ReRAM Model (https://eprints.soton.ac.uk/411693/).

    Parameters
    ----------
    time_series_resolution : float
        Time series resolution (s).
    r_off : float
        Off (maximum) resistance of the device (ohms).
    r_on : float
        On (minimum) resistance of the device (ohms).
    A_p : float
        A_p model parameter.
    A_n : float
        A_n model parameter.
    t_p : float
        t_p model parameter.
    t_n : float
        t_n model parameter.
    k_p : float
        k_p model parameter.
    k_n : float
        k_n model parameter.
    r_p : float
        r_p voltage-dependent resistive boundary function coefficients.
    r_n : float
        r_n voltage-dependent resistive boundary function coefficients.
    eta : int
        Switching direction to stimulus polarity.
    a_p : float
        a_p model parameter.
    a_n : float
        a_n model parameter.
    b_p : float
        b_p model parameter.
    b_n : float
        b_n model parameter.
    """

    def __init__(
        self,
        time_series_resolution=1e-8,
        r_off=17e3,
        r_on=1280,
        A_p=743.47,
        A_n=-68012.28374,
        t_p=6.51,
        t_n=0.31645,
        k_p=5.11e-4,
        k_n=1.17e-3,
        r_p=[16719, 0],
        r_n=[29304.82557, 23692.77225],
        eta=1,
        a_p=0.24,
        a_n=0.24,
        b_p=3,
        b_n=3,
        **kwargs
    ):

        args = memtorch.bh.unpack_parameters(locals())
        super(Data_Driven, self).__init__(
            args.r_off, args.r_on, args.time_series_resolution, 0, 0
        )
        self.A_p = args.A_p
        self.A_n = args.A_n
        self.t_p = args.t_p
        self.t_n = args.t_n
        self.k_p = args.k_p
        self.k_n = args.k_n
        self.r_p = args.r_p
        self.r_n = args.r_n
        self.eta = args.eta
        self.a_p = args.a_p
        self.a_n = args.a_n
        self.b_p = args.b_p
        self.b_n = args.b_n
        self.g = 1 / self.r_on

    def simulate(self, voltage_signal, return_current=False):
        len_voltage_signal = 1
        try:
            len_voltage_signal = len(voltage_signal)
        except:
            voltage_signal = [voltage_signal]

        if return_current:
            current = np.zeros(len_voltage_signal)

        np.seterr(all="raise")
        for t in range(0, len_voltage_signal):
            current_ = self.current(voltage_signal[t])
            self.g = 1 / self.resistance(voltage_signal[t])
            if return_current:
                current[t] = current_

        if return_current:
            return current

    def current(self, voltage):
        """Method to determine the current of the model given an applied voltage.

        Parameters
        ----------
        voltage : float
            The current applied voltage (V).

        Returns
        -------
        float
            The observed current (A).
        """
        if voltage > 0:
            return self.a_p * self.g * math.sinh(self.b_p * voltage)
        else:
            return self.a_n * self.g * math.sinh(self.b_n * voltage)

    def resistance(self, voltage):
        """Method to determine the resistance of the model given an applied voltage.

        Parameters
        ----------
        voltage : float
            The current applied voltage (V).

        Returns
        -------
        float
            The observed resistance (Ω).
        """

        def r_pn(voltage, r_pn):
            sum = 0
            for m_pn in range(0, len(r_pn)):
                sum += r_pn[m_pn] * (voltage ** (m_pn))

            return sum

        if voltage > 0:
            r_pn_eval = r_pn(voltage, self.r_p)
            resistance_ = (
                np.log(
                    np.exp(self.eta * self.k_p * r_pn_eval)
                    + np.exp(
                        -self.eta
                        * self.k_p
                        * (self.A_p * (math.exp(self.t_p * abs(voltage)) - 1))
                        * self.time_series_resolution
                    )
                    * (
                        np.exp(self.eta * self.k_p * (1 / self.g))
                        - np.exp(self.eta * self.k_p * r_pn_eval)
                    )
                )
                / self.k_p
            )
            if resistance_ > self.eta * r_pn_eval:
                return 1 / self.g
            else:
                return max(
                    min(resistance_, self.r_off), self.r_on
                )  # Artificially confine the resistance between r_on and r_off
        elif voltage < 0:
            r_pn_eval = r_pn(voltage, self.r_n)
            resistance_ = (
                -np.log(
                    np.exp(
                        -self.eta * self.k_n * (1 / self.g)
                        + self.eta
                        * self.k_n
                        * (self.A_n * (-1 + np.exp(self.t_n * abs(voltage))))
                        * self.time_series_resolution
                    )
                    - np.exp(-self.eta * self.k_n * r_pn_eval)
                    * (
                        -1
                        + np.exp(
                            self.eta
                            * self.k_n
                            * (self.A_n * (-1 + np.exp(self.t_n * abs(voltage))))
                            * self.time_series_resolution
                        )
                    )
                )
                / self.k_n
            )
            if resistance_ < self.eta * r_pn_eval:
                return 1 / self.g
            else:
                return max(
                    min(resistance_, self.r_off), self.r_on
                )  # Artificially confine the resistance between r_on and r_off
        else:
            return 1 / self.g

    def set_conductance(self, conductance):
        conductance = clip(conductance, 1 / self.r_off, 1 / self.r_on)
        self.g = conductance

    def plot_hysteresis_loop(
        self,
        duration=1e-3,
        voltage_signal_amplitude=1.5,
        voltage_signal_frequency=10e3,
        return_result=False,
    ):
        return super().plot_hysteresis_loop(
            self,
            duration=duration,
            voltage_signal_amplitude=voltage_signal_amplitude,
            voltage_signal_frequency=voltage_signal_frequency,
            return_result=return_result,
        )

    def plot_bipolar_switching_behaviour(
        self,
        voltage_signal_amplitude=1.5,
        voltage_signal_frequency=10e3,
        log_scale=True,
        return_result=False,
    ):
        return super().plot_bipolar_switching_behaviour(
            self,
            voltage_signal_amplitude=voltage_signal_amplitude,
            voltage_signal_frequency=voltage_signal_frequency,
            log_scale=log_scale,
            return_result=return_result,
        )



import numpy as np
import torch
import torch.functional as F
import torch.nn as nn
from sklearn import datasets, linear_model
from sklearn.metrics import r2_score

import memtorch
from memtorch.utils import pad_tensor


def naive_tune(module, input_shape, verbose=True):
    """Method to determine a linear relationship between a memristive crossbar and the output for a given memristive module.

    Parameters
    ----------
    module : torch.nn.Module
        Memristive layer to tune.
    input_shape : int, int
        Shape of the randomly generated input used to tune a crossbar.
    verbose : bool, optional
        Used to determine if verbose output is enabled (True) or disabled (False).

    Returns
    -------
    function
        Function which transforms the output of the crossbar to the expected output.
    """
    device = torch.device("cpu" if "cpu" in memtorch.__version__ else "cuda")
    tmp = module.bias
    module.bias = None
    input = torch.rand(input_shape).uniform_(-1, 1).to(device)
    initial_forward_legacy_state = module.forward_legacy_enabled
    module.forward_legacy_enabled = False
    output = module.forward(input).detach().cpu()
    module.forward_legacy_enabled = True
    legacy_output = module.forward(input).detach().cpu()
    module.forward_legacy_enabled = initial_forward_legacy_state
    output = output.numpy().reshape(-1, 1)
    legacy_output = legacy_output.numpy().reshape(-1, 1)
    reg = linear_model.LinearRegression(fit_intercept=True).fit(output, legacy_output)
    coef = np.array(reg.coef_).item()
    intercept = np.array(reg.intercept_).item()

    def transform_output(x):
        return x * coef + intercept

    module.bias = tmp
    if verbose:
        print(
            "Tuned %s. Coefficient of determination: %f [%f, %f]"
            % (module, reg.score(output, legacy_output), coef, intercept)
        )

    return transform_output



import itertools
import math
import multiprocessing
from enum import Enum, auto, unique

import numpy as np
import pandas as pd
import torch
import torch.multiprocessing as mp
import torch.nn as nn

import memtorch
from memtorch.bh.memristor import Data_Driven2021

if "cpu" not in memtorch.__version__:
    import memtorch_cuda_bindings

from .Tile import gen_tiles

CUDA_supported_memristor_models = [Data_Driven2021]


@unique
class Scheme(Enum):
    """Scheme enumeration."""

    SingleColumn = auto()
    DoubleColumn = auto()


class Crossbar:
    """Class used to model memristor crossbars.

    Parameters
    ----------
    memristor_model : memtorch.bh.memristor.Memristor.Memristor
        Memristor model.
    memristor_model_params: **kwargs
        **kwargs to instantiate the memristor model with.
    shape : int, int
        Shape of the crossbar.
    tile_shape : int, int
        Tile shape to use to store weights. If None, modular tiles are not used.
    use_bindings : bool
        Used to determine if C++/CUDA bindings are used (True) or not (False).
    random_crossbar_init: bool
        Determines if the crossbar is to be initialized at random values in between Ron and Roff
    """

    def __init__(
        self,
        memristor_model,
        memristor_model_params,
        shape,
        tile_shape=None,
        use_bindings=True,
        cuda_malloc_heap_size=50,
        random_crossbar_init=False,
    ):
        self.memristor_model_params = memristor_model_params
        self.time_series_resolution = memristor_model_params.get(
            "time_series_resolution"
        )
        self.device = torch.device("cpu" if "cpu" in memtorch.__version__ else "cuda")
        self.tile_shape = tile_shape
        self.use_bindings = use_bindings
        self.cuda_malloc_heap_size = cuda_malloc_heap_size
        if hasattr(memristor_model_params, "r_off"):
            self.r_off_mean = memristor_model_params["r_off"]
            if callable(self.r_off_mean):
                self.r_off_mean = self.r_off_mean()
        else:
            self.r_off_mean = memristor_model().r_off

        if hasattr(memristor_model_params, "r_on"):
            self.r_on_mean = memristor_model_params["r_on"]
            if callable(self.r_on_mean):
                self.r_on_mean = self.r_on_mean()
        else:
            self.r_on_mean = memristor_model().r_on

        if len(shape) == 4:  # memtorch.mn.Conv2d and memtorch.mn.Conv3d
            self.rows = shape[1] * shape[2] * shape[3]
            self.columns = shape[0]
        elif len(shape) == 3:  # memtorch.mn.Conv1d
            self.rows = shape[1] * shape[2]
            self.columns = shape[0]
        elif len(shape) == 2:  # memtorch.mn.Linear
            self.columns, self.rows = shape
        else:
            raise Exception("Unsupported crossbar shape.")

        self.rows = int(self.rows)
        self.columns = int(self.columns)
        if tile_shape is not None:
            self.tiles_map = None
            tiles_num = math.ceil(self.rows / tile_shape[0]) * math.ceil(
                self.columns / tile_shape[1]
            )
            self.devices = np.empty(
                (tiles_num, tile_shape[0], tile_shape[1]), dtype=object
            )
            self.devices.flat = [
                memristor_model(**memristor_model_params) for _ in self.devices.flat
            ]
            if random_crossbar_init:
                self.conductance_matrix = torch.FloatTensor(
                    np.random.uniform(
                        1 / self.r_off_mean,
                        1 / self.r_on_mean,
                        size=(tiles_num, tile_shape[0], tile_shape[1]),
                    )
                ).to(self.device)
            else:
                self.conductance_matrix = torch.zeros(
                    (tiles_num, tile_shape[0], tile_shape[1])
                )
        else:
            self.devices = np.empty((self.rows, self.columns), dtype=object)
            self.devices.flat = [
                memristor_model(**memristor_model_params) for _ in self.devices.flat
            ]
            if random_crossbar_init:
                self.conductance_matrix = torch.FloatTensor(
                    np.random.uniform(
                        1 / self.r_off_mean,
                        1 / self.r_on_mean,
                        size=(self.rows, self.columns),
                    )
                ).to(self.device)
            else:
                self.conductance_matrix = torch.zeros((self.rows, self.columns)).to(
                    self.device
                )

        self.g_np = np.vectorize(lambda x: x.g)
        self.update(from_devices=False)
        self.max_abs_conductance = torch.abs(self.conductance_matrix).flatten()

    def update(self, from_devices=True, parallelize=False):
        """Method to update either the layers conductance_matrix or each devices conductance state.

        Parameters
        ----------
        from_devices : bool
            The conductance matrix can either be updated from all devices (True), or each device can be updated from the conductance matrix (False).
        parallelize : bool
            The operation is parallelized (True).
        """
        if from_devices:
            self.conductance_matrix = torch.tensor(self.g_np(self.devices)).to(
                self.device
            )
            self.max_abs_conductance = (
                torch.abs(self.conductance_matrix).flatten().max()
            )
        else:
            if parallelize:

                def write_conductance(device, conductance):
                    device.set_conductance(conductance)

                np.frompyfunc(write_conductance, 2, 0)(
                    self.devices, self.conductance_matrix.detach().cpu()
                )
            else:
                if self.tile_shape is not None:
                    for i in range(0, self.devices.shape[0]):
                        for j in range(0, self.devices.shape[1]):
                            for k in range(0, self.devices.shape[2]):
                                self.devices[i][j][k].set_conductance(
                                    self.conductance_matrix[i][j][k].item()
                                )
                else:
                    for i in range(0, self.rows):
                        for j in range(0, self.columns):
                            self.devices[i][j].set_conductance(
                                self.conductance_matrix[i][j].item()
                            )

    def write_conductance_matrix(
        self,
        conductance_matrix,
        transistor=True,
        programming_routine=None,
        programming_routine_params={},
    ):
        """Method to directly program (alter) the conductance of all devices within the crossbar.

        Parameters
        ----------
        conductance_matrix : torch.FloatTensor
            Conductance matrix to write.
        transistor : bool
            Used to determine if a 1T1R (True) or 0T1R arrangement (False) is simulated.
        programming_routine
            Programming routine (method) to use.
        programming_routine_params : **kwargs
            Programming routine keyword arguments.
        """
        if (
            len(conductance_matrix.shape) == 3 or len(conductance_matrix.shape) == 4
        ):  # memtorch.mn.Conv1d, memtorch.mn.Conv2d, and memtorch.mn.Conv3d
            conductance_matrix = conductance_matrix.reshape(self.columns, self.rows).T
        elif len(conductance_matrix.shape) == 2:  # memtorch.mn.Linear
            conductance_matrix = conductance_matrix.T.clone().detach().to(self.device)
            assert (
                conductance_matrix.shape[0] == self.rows
                and conductance_matrix.shape[1] == self.columns
            )
        else:
            raise Exception("Unsupported crossbar shape.")

        if self.tile_shape is not None:
            conductance_matrix, tiles_map = gen_tiles(
                conductance_matrix,
                self.tile_shape,
                input=False,
                use_bindings=self.use_bindings,
            )
            self.tiles_map = tiles_map

        min = (
            torch.tensor(1 / np.vectorize(lambda x: x.r_off)(self.devices))
            .to(self.device)
            .float()
        )
        max = (
            torch.tensor(1 / np.vectorize(lambda x: x.r_on)(self.devices))
            .to(self.device)
            .float()
        )
        conductance_matrix = torch.max(
            torch.min(conductance_matrix.to(self.device), max), min
        )
        if transistor:
            self.conductance_matrix = conductance_matrix
            self.max_abs_conductance = (
                torch.abs(self.conductance_matrix).flatten().max()
            )
            self.update(from_devices=False)
        else:
            if (
                self.use_bindings
                and type(self.devices.any()) in CUDA_supported_memristor_models
                and "cpu" not in memtorch.__version__
            ):
                device_matrix = torch.FloatTensor(self.g_np(self.devices))
                device_matrix_aug = device_matrix
                conductance_matrix_aug = conductance_matrix
                if (
                    len(device_matrix.shape) == 2
                ):  # To ensure compatibility with CUDA code
                    device_matrix_aug = device_matrix[:, :, None]
                    conductance_matrix_aug = conductance_matrix[:, :, None]

                self.conductance_matrix = memtorch_cuda_bindings.simulate_passive(
                    conductance_matrix_aug,
                    device_matrix_aug,
                    self.cuda_malloc_heap_size,
                    **programming_routine_params,
                    **self.memristor_model_params
                )
                self.max_abs_conductance = (
                    torch.abs(self.conductance_matrix).flatten().max()
                )
                self.update(from_devices=False)
            else:
                assert (
                    programming_routine is not None
                ), "If memtorch_cuda_bindings.simulate_passive is not used, a programming routine must be provided."
                if self.tile_shape is not None:
                    for i in range(0, self.devices.shape[0]):
                        for j in range(0, self.devices.shape[1]):
                            for k in range(0, self.devices.shape[2]):
                                self.devices = programming_routine(
                                    self,
                                    (i, j, k),
                                    conductance_matrix[i][j][k],
                                    **programming_routine_params
                                )
                else:
                    for i in range(0, self.rows):
                        for j in range(0, self.columns):
                            self.devices = programming_routine(
                                self,
                                (i, j),
                                conductance_matrix[i][j],
                                **programming_routine_params
                            )

                self.update(from_devices=True)


def init_crossbar(
    weights,
    memristor_model,
    memristor_model_params,
    transistor,
    mapping_routine,
    programming_routine,
    programming_routine_params={},
    p_l=None,
    scheme=Scheme.DoubleColumn,
    tile_shape=(128, 128),
    use_bindings=True,
    cuda_malloc_heap_size=50,
    random_crossbar_init=False,
):
    """Method to initialise and construct memristive crossbars.

    Parameters
    ----------
    weights : torch.Tensor
        Weights to map.
    memristor_model : memtorch.bh.memristor.Memristor.Memristor
        Memristor model.
    memristor_model_params: **kwargs
        **kwargs to instantiate the memristor model with.
    transistor : bool
        Used to determine if a 1T1R (True) or 1R arrangement (False) is simulated.
    mapping_routine : function
        Mapping routine to use.
    programming_routine : function
        Programming routine to use.
    programming_routine_params : **kwargs
        Programming routine keyword arguments.
    p_l: float
        If not None, the proportion of weights to retain.
    scheme : memtorch.bh.Scheme
        Scheme enum.
    tile_shape : int, int
        Tile shape to use to store weights. If None, modular tiles are not used.
    use_bindings : bool
        Used to determine if C++/CUDA bindings are used (True) or not (False).
    random_crossbar_init: boolean
        Determines if the crossbar is to be initialized at random values in between Ron and Roff

    Returns
    -------
    tuple
        The constructed crossbars and forward() function.
    """
    assert scheme in Scheme, "scheme must be a Scheme Enum."
    weights_ = weights.data.detach().clone()
    crossbars = []
    reference_memristor_model_params = {**memristor_model_params, **{"reference": True}}
    reference_memristor_model = memristor_model(**reference_memristor_model_params)
    if scheme == Scheme.DoubleColumn:
        if len(weights.shape) == 5:  # memtorch.mn.Conv3d
            channel_idx = 0
            for channel in range(weights.shape[1]):
                channel_weights = weights.detach().clone()[:, channel, :, :, :]
                crossbars.append(
                    memtorch.bh.crossbar.Crossbar(
                        memristor_model,
                        memristor_model_params,
                        channel_weights.shape,
                        tile_shape,
                        use_bindings=use_bindings,
                        cuda_malloc_heap_size=cuda_malloc_heap_size,
                        random_crossbar_init=random_crossbar_init,
                    )
                )
                crossbars.append(
                    memtorch.bh.crossbar.Crossbar(
                        memristor_model,
                        memristor_model_params,
                        channel_weights.shape,
                        tile_shape,
                        use_bindings=use_bindings,
                        cuda_malloc_heap_size=cuda_malloc_heap_size,
                        random_crossbar_init=random_crossbar_init,
                    )
                )
                pos_conductance_matrix, neg_conductance_matrix = mapping_routine(
                    channel_weights,
                    reference_memristor_model.r_on,
                    reference_memristor_model.r_off,
                    scheme=scheme,
                    p_l=p_l,
                )
                crossbars[channel_idx].write_conductance_matrix(
                    pos_conductance_matrix,
                    transistor=transistor,
                    programming_routine=programming_routine,
                    programming_routine_params=programming_routine_params,
                )
                crossbars[channel_idx + 1].write_conductance_matrix(
                    neg_conductance_matrix,
                    transistor=transistor,
                    programming_routine=programming_routine,
                    programming_routine_params=programming_routine_params,
                )
                channel_idx += 2
        else:
            crossbars.append(
                memtorch.bh.crossbar.Crossbar(
                    memristor_model,
                    memristor_model_params,
                    weights.shape,
                    tile_shape,
                    use_bindings=use_bindings,
                    random_crossbar_init=random_crossbar_init,
                )
            )
            crossbars.append(
                memtorch.bh.crossbar.Crossbar(
                    memristor_model,
                    memristor_model_params,
                    weights.shape,
                    tile_shape,
                    use_bindings=use_bindings,
                    random_crossbar_init=random_crossbar_init,
                )
            )
            pos_conductance_matrix, neg_conductance_matrix = mapping_routine(
                weights_,
                reference_memristor_model.r_on,
                reference_memristor_model.r_off,
                scheme=scheme,
                p_l=p_l,
            )
            crossbars[0].write_conductance_matrix(
                pos_conductance_matrix,
                transistor=transistor,
                programming_routine=programming_routine,
                programming_routine_params=programming_routine_params,
            )
            crossbars[1].write_conductance_matrix(
                neg_conductance_matrix,
                transistor=transistor,
                programming_routine=programming_routine,
                programming_routine_params=programming_routine_params,
            )

        def out(crossbars, operation, idx=(0, 1), **kwargs):
            assert (
                len(idx) == 2
            ), "idx must contain indicies of the positive and negative crossbars"
            return operation(crossbars[idx[0]], **kwargs) - operation(
                crossbars[idx[1]], **kwargs
            )

    elif scheme == Scheme.SingleColumn:
        if len(weights.shape) == 5:  # memtorch.mn.Conv3d
            channel_idx = 0
            for channel in range(weights.shape[1]):
                channel_weights = weights.detach().clone()[:, channel, :, :, :]
                crossbars.append(
                    memtorch.bh.crossbar.Crossbar(
                        memristor_model,
                        memristor_model_params,
                        channel_weights.shape,
                        tile_shape,
                        use_bindings=use_bindings,
                        random_crossbar_init=random_crossbar_init,
                    )
                )
                conductance_matrix = mapping_routine(
                    channel_weights,
                    reference_memristor_model.r_on,
                    reference_memristor_model.r_off,
                    scheme=scheme,
                    p_l=p_l,
                )
                crossbars[channel_idx].write_conductance_matrix(
                    conductance_matrix,
                    transistor=transistor,
                    programming_routine=programming_routine,
                    programming_routine_params=programming_routine_params,
                )
                channel_idx += 1
        else:
            crossbars.append(
                memtorch.bh.crossbar.Crossbar(
                    memristor_model,
                    memristor_model_params,
                    weights.shape,
                    tile_shape,
                    use_bindings=use_bindings,
                    random_crossbar_init=random_crossbar_init,
                )
            )
            conductance_matrix = mapping_routine(
                weights_,
                reference_memristor_model.r_on,
                reference_memristor_model.r_off,
                scheme=scheme,
                p_l=p_l,
            )
            crossbars[0].write_conductance_matrix(
                conductance_matrix,
                transistor=transistor,
                programming_routine=programming_routine,
                programming_routine_params=programming_routine_params,
            )

        g_m = (
            (1 / reference_memristor_model.r_on) + (1 / reference_memristor_model.r_off)
        ) / 2

        def out(crossbars, operation, idx=0, **kwargs):
            return operation(crossbars[idx], **kwargs) - g_m

    else:
        raise ("%s is not currently supported." % scheme)

    return crossbars, out


def simulate_matmul(
    input,
    crossbar,
    nl=True,
    tiles_map=None,
    crossbar_shape=None,
    max_input_voltage=None,
    ADC_resolution=None,
    ADC_overflow_rate=0.0,
    quant_method=None,
    use_bindings=True,
):
    """Method to simulate non-linear IV device characterisitcs for a 2-D crossbar architecture given scaled inputs.

    Parameters
    ----------
    input : torch.Tensor
        Scaled input tensor.
    crossbar : memtorch.bh.Crossbar
        Crossbar containing devices to simulate.
    nl : bool
        Use lookup tables rather than simulating each device (True).
    tiles_map: torch.Tensor
        Tiles map for devices if tile_shape is not None.
    crossbar_shape : int, int
        Crossbar shape if tile_shape is not None.
    max_input_voltage : float
        Maximum input voltage used to encode inputs. If None, inputs are unbounded.
    ADC_resolution : int
        ADC resolution (bit width). If None, quantization noise is not accounted for.
    ADC_overflow_rate : float
        Overflow rate threshold for linear quanitzation (if ADC_resolution is not None).
    quant_method:
        Quantization method. Must be in memtorch.bh.Quantize.quant_methods.
    use_bindings : bool
        Used to determine if C++/CUDA bindings are used (True) or not (False).

    Returns
    -------
    torch.Tensor
        Output tensor.
    """
    devices = crossbar.devices
    if max_input_voltage is not None:
        output_max = crossbar.max_abs_conductance * max_input_voltage
    else:
        output_max = float("inf")

    del crossbar
    assert len(devices.shape) == 2 or len(devices.shape) == 3, "Invalid devices shape."
    device = torch.device("cpu" if "cpu" in memtorch.__version__ else "cuda")
    if quant_method is not None:
        assert (
            ADC_resolution is not None
            and type(ADC_resolution) == int
            and ADC_resolution > 0
        ), "ADC resolution is invalid."
        assert (
            quant_method in memtorch.bh.Quantize.quant_methods
        ), "quant_method is not valid."
        assert (
            ADC_overflow_rate is not None
        ), "ADC_overflow_rate must be specified if quant_method is not None."

    input_rows, input_columns = input.shape
    if len(devices.shape) == 2:
        devices_rows, devices_columns = devices.shape
        mat_res_ = torch.zeros((input_rows, devices_columns)).to(device)
        if nl:
            for i in range(input_rows):
                for j in range(devices_columns):
                    for k in range(input_columns):
                        mat_res_[i][j] += devices[k][j].g * input[i][k]
        else:
            for i in range(input_rows):
                for j in range(devices_columns):
                    for k in range(input_columns):
                        mat_res_[i][j] += (
                            devices[k][j]
                            .simulate(
                                torch.Tensor([input[i][k]]).cpu(), return_current=True
                            )
                            .item()
                        )

        mat_res_ = torch.clamp(mat_res_, min=-output_max, max=output_max)
        if quant_method is not None:
            mat_res_ = memtorch.bh.Quantize.quantize(
                mat_res_,
                quant=ADC_resolution,
                overflow_rate=ADC_overflow_rate,
                quant_method=quant_method,
            )
    else:
        assert (
            tiles_map is not None and crossbar_shape is not None
        ), "tiles_map is not None."
        tile_shape = devices.shape[-2:]
        input_tiles, input_tiles_map = gen_tiles(
            input, tile_shape, input=True, use_bindings=use_bindings
        )
        mat_res_ = torch.zeros((input.shape[0], crossbar_shape[1])).to(device)

        def tile_simulate_matmul_row(
            input_row_tiles,
            input_tiles_map,
            devices,
            tiles_map,
            crossbar_shape,
            nl,
            ADC_resolution,
            ADC_overflow_rate,
            quant_method,
        ):
            device = torch.device("cpu" if "cpu" in memtorch.__version__ else "cuda")
            tile_shape = devices.shape[-2:]
            partial_sum = torch.zeros((tiles_map.shape[1], tile_shape[1])).to(device)
            for j in range(tiles_map.shape[1]):
                for i in range(tiles_map.shape[0]):
                    tile_a = input_row_tiles[int(input_tiles_map[i])]
                    if len(tile_a.shape) == 1:
                        tile_a = tile_a.unsqueeze(0)

                    tile_b = devices[int(tiles_map[i][j])]
                    mat_res = torch.zeros((tile_a.shape[0], tile_b.shape[1])).to(device)
                    for ii in range(tile_a.shape[0]):
                        for jj in range(tile_b.shape[1]):
                            for kk in range(tile_b.shape[0]):
                                if nl:
                                    mat_res[ii][jj] += (
                                        tile_a[ii][kk].item() * tile_b[kk][jj].g
                                    )
                                else:
                                    mat_res[ii][jj] += (
                                        tile_b[kk][jj]
                                        .simulate(
                                            torch.Tensor([tile_a[ii][kk]]).cpu(),
                                            return_current=True,
                                        )
                                        .item()
                                    )

                    mat_res = torch.clamp(mat_res, min=-output_max, max=output_max)
                    if quant_method is not None:
                        partial_sum[j] += memtorch.bh.Quantize.quantize(
                            mat_res.squeeze(),
                            quant=ADC_resolution,
                            overflow_rate=ADC_overflow_rate,
                            quant_method=quant_method,
                        )
                    else:
                        partial_sum[j] += mat_res.squeeze()

            output_act = partial_sum.flatten()
            output_act = output_act[: crossbar_shape[1]]
            return output_act

        if input_tiles.shape[-2] > 1:
            for row_idx in range(input_tiles.shape[-2]):
                mat_res_[row_idx] = tile_simulate_matmul_row(
                    input_tiles[:, row_idx, :],
                    input_tiles_map,
                    devices,
                    tiles_map,
                    crossbar_shape,
                    nl,
                    ADC_resolution,
                    ADC_overflow_rate,
                    quant_method,
                )
        else:
            mat_res_ = tile_simulate_matmul_row(
                input_tiles,
                input_tiles_map,
                devices,
                tiles_map,
                crossbar_shape,
                nl,
                ADC_resolution,
                ADC_overflow_rate,
                quant_method,
            )

    return mat_res_



# Modular tile implementation based on: https://github.com/xxwang1/DNN-accelerator-based-on-tiled-architecture
import math

import numpy as np
import torch
import torch.nn as nn

import memtorch

if "cpu" in memtorch.__version__:
    import memtorch_bindings
else:
    import memtorch_cuda_bindings as memtorch_bindings


class Tile:
    """Class used to create modular crossbar tiles to represent 2D matrices.

    Parameters
    ----------
    tile_shape : int, int
        Tile shape to use to store weights.
    patch_num : int
        Patch number.
    """

    def __init__(self, tile_shape, patch_num=None):
        self.tile_shape = tile_shape
        self.patch_num = patch_num
        if patch_num is None:
            self.array = torch.zeros(tile_shape)
        else:
            self.array = torch.zeros((patch_num, tile_shape[0]))

    def update_array(self, new_array):
        """Method to update the tile's weights.

        Parameters
        ----------
        new_array : torch.Tensor
            New array to construct the tile with.
        """
        if new_array.shape == self.tile_shape or new_array.shape == (
            self.patch_num,
            self.tile_shape[0],
        ):
            self.array = new_array
        else:
            new_col_cnt = new_array.shape[1]
            if type(new_array) == np.ndarray:
                new_array = torch.from_numpy(new_array)
            else:
                new_array = new_array.clone().detach()

            if self.patch_num is None:
                new_row_cnt = new_array.shape[0]
                self.array[:new_row_cnt, :new_col_cnt] = new_array
            else:
                self.array[:, :new_col_cnt] = new_array


def gen_tiles(tensor, tile_shape, input=False, use_bindings=True):
    """Method to generate a set of modular tiles representative of a tensor.

    Parameters
    ----------
    tensor : torch.Tensor
        Tensor to represent using modular crossbar tiles.
    tile_shape : int, int
        Tile shape to use to store weights.
    input : bool
        Used to determine if a tensor is an input (True).

    Returns
    -------
    torch.Tensor, torch.Tensor
        Tiles and tile_map.
    """
    if use_bindings:
        tiles, tiles_map = memtorch_bindings.gen_tiles(tensor, tile_shape, input)
        return tiles, tiles_map
    else:
        tiles = []
        tensor_shape = tensor.shape
        if input:
            patch_num = tensor_shape[0]
            tile_columns = math.ceil(
                tensor_shape[1] / tile_shape[0]
            )  # Number of mapped arrays
            tiles_map = torch.empty([tile_columns])
            for tile_column in range(tile_columns):
                tiles.append(Tile(patch_num=patch_num, tile_shape=tile_shape))
                column_start = (
                    tile_column * tile_shape[0]
                )  # Set the range of the array slice by defining starting and ending columns
                if tile_column == tile_columns - 1:  # Execute if last column
                    column_end = -1
                else:
                    column_end = (tile_column + 1) * tile_shape[0]

                if column_end == -1:  # If the last column
                    tiles[-1].update_array(tensor[:, column_start:])
                else:
                    tiles[-1].update_array(tensor[:, column_start:(column_end)])

                new_tile_id = len(tiles) - 1
                tiles_map[tile_column] = new_tile_id
        else:
            tile_rows = math.ceil(tensor_shape[0] / tile_shape[0])
            tile_columns = math.ceil(tensor_shape[1] / tile_shape[1])
            tiles_map = torch.empty([tile_rows, tile_columns])
            for tile_row in range(tile_rows):
                row_start = tile_row * tile_shape[0]
                if tile_row == tile_rows - 1:  # Execute if last row
                    row_end = -1
                else:
                    row_end = (tile_row + 1) * tile_shape[0]

                for tile_column in range(tile_columns):
                    tiles.append(Tile(tile_shape=tile_shape))
                    column_start = (
                        tile_column * tile_shape[1]
                    )  # Set the range of the array slice by defining starting and ending columns
                    if tile_column == tile_columns - 1:  # Execute if last column
                        column_end = -1
                    else:
                        column_end = (tile_column + 1) * tile_shape[1]

                    if (
                        row_end == -1 and column_end == -1
                    ):  # If last row and last column
                        tiles[-1].update_array(tensor[row_start:, column_start:])
                    elif (
                        row_end == -1 and column_end != -1
                    ):  # If last row but not last column
                        tiles[-1].update_array(
                            tensor[row_start:, column_start:column_end]
                        )
                    elif (
                        row_end != -1 and column_end == -1
                    ):  # If last column but not last row
                        tiles[-1].update_array(tensor[row_start:row_end, column_start:])
                    else:  # If neither last row nor last column
                        tiles[-1].update_array(
                            tensor[row_start:(row_end), column_start:(column_end)]
                        )

                    new_tile_id = len(tiles) - 1
                    tiles_map[tile_row][tile_column] = new_tile_id

        tiles = torch.tensor([np.array(tile.array.detach().cpu()) for tile in tiles])
    return tiles, tiles_map


def tile_matmul_row(
    mat_a_row_tiles,
    mat_a_tiles_map,
    mat_b_tiles,
    mat_b_tiles_map,
    mat_b_shape,
    source_resistance=None,
    line_resistance=None,
    ADC_resolution=None,
    ADC_overflow_rate=0.0,
    quant_method=None,
    transistor=True,
):
    """Method to perform row-wise tile matrix multiplication, given two sets of tiles, using a pythonic approach.

    Parameters
    ----------
    mat_a_row_tiles : torch.Tensor
        Tiles representing a row of matrix A.
    mat_a_tiles_map : torch.Tensor
        Tiles map for matrix A.
    mat_b_tiles : torch.Tensor
        Tiles representing matrix B.
    mat_b_tiles_map : torch.Tensor
        Tiles map for matrix B.
    mat_b_shape : int, int
        Shape of matrix B.
    source_resistance : float
        The resistance between word/bit line voltage sources and crossbar(s).
    line_resistance : float
        The interconnect line resistance between adjacent cells.
    ADC_resolution : int
        ADC resolution (bit width). If None, quantization noise is not accounted for.
    ADC_overflow_rate : float
        Overflow rate threshold for linear quanitzation (if ADC_resolution is not None).
    quant_method: str
        Quantization method. Must be in memtorch.bh.Quantize.quant_methods.
    transistor : bool
        TBD.

    Returns
    -------
    torch.Tensor
        Output tensor.
    """
    device = torch.device("cpu" if "cpu" in memtorch.__version__ else "cuda")
    if quant_method is not None:
        assert (
            ADC_resolution is not None
            and type(ADC_resolution) == int
            and ADC_resolution > 0
        ), "ADC resolution is invalid."
        assert (
            quant_method in memtorch.bh.Quantize.quant_methods
        ), "quant_method is not valid."
        assert (
            ADC_overflow_rate is not None
        ), "ADC_overflow_rate must be specified if quant_method is not None."

    tile_shape = mat_b_tiles.shape[-2:]
    partial_sum = torch.zeros((mat_b_tiles_map.shape[1], tile_shape[1])).to(device)
    for j in range(mat_b_tiles_map.shape[1]):
        for i in range(mat_b_tiles_map.shape[0]):
            tile_a = mat_a_row_tiles[int(mat_a_tiles_map[i])]
            tile_b = mat_b_tiles[int(mat_b_tiles_map[i][j])]
            if transistor:
                if quant_method is not None:
                    partial_sum[j] += memtorch.bh.Quantize.quantize(
                        torch.matmul(tile_a.to(device), tile_b.to(device)).squeeze(),
                        quant=ADC_resolution,
                        overflow_rate=ADC_overflow_rate,
                        quant_method=quant_method,
                    )
                else:
                    partial_sum[j] += torch.matmul(
                        tile_a.to(device), tile_b.to(device)
                    ).squeeze()
            else:
                if quant_method is not None:
                    partial_sum[j] += memtorch.bh.crossbar.Passive.solve_passive(
                        tile_b,
                        tile_a,
                        torch.zeros(tile_b.shape[1]),
                        source_resistance,
                        line_resistance,
                        det_readout_currents=True,
                        use_bindings=False,
                    )
                else:
                    partial_sum[j] += memtorch.bh.Quantize.quantize(
                        memtorch.bh.crossbar.Passive.solve_passive(
                            tile_b,
                            tile_a,
                            torch.zeros(tile_b.shape[1]),
                            source_resistance,
                            line_resistance,
                            det_readout_currents=True,
                            use_bindings=False,
                        ),
                        quant=ADC_resolution,
                        overflow_rate=ADC_overflow_rate,
                        quant_method=quant_method,
                    )

    output_act = partial_sum.flatten()
    output_act = output_act[: mat_b_shape[1]]
    return output_act


def tile_matmul(
    mat_a_tiles,
    mat_a_tiles_map,
    mat_a_shape,
    mat_b_tiles,
    mat_b_tiles_map,
    mat_b_shape,
    source_resistance=None,
    line_resistance=None,
    ADC_resolution=None,
    ADC_overflow_rate=0.0,
    quant_method=None,
    transistor=True,
    use_bindings=True,
    cuda_malloc_heap_size=50,
):
    """Method to perform 2D matrix multiplication, given two sets of tiles.

    Parameters
    ----------
    mat_a_tiles : torch.Tensor
        Tiles representing matrix A.
    mat_a_tiles_map : torch.Tensor
        Tiles map for matrix A.
    mat_a_shape : int, int
        Shape of matrix A.
    mat_b_tiles : torch.Tensor
        Tiles representing matrix B.
    mat_b_tiles_map : torch.Tensor
        Tiles map for matrix B.
    mat_b_shape : int, int
        Shape of matrix B.
    source_resistance : float
        The resistance between word/bit line voltage sources and crossbar(s).
    line_resistance : float
        The interconnect line resistance between adjacent cells.
    ADC_resolution : int
        ADC resolution (bit width). If None, quantization noise is not accounted for.
    ADC_overflow_rate : float
        Overflow rate threshold for linear quanitzation (if ADC_resolution is not None).
    quant_method: str
        Quantization method. Must be in memtorch.bh.Quantize.quant_methods.
    transistor : bool
        TBD.
    use_bindings : bool
        Use C++/CUDA bindings to parallelize tile_matmul operations (True).
    cuda_malloc_heap_size : int
        cudaLimitMallocHeapSize (in MB) to determine allocatable kernel heap memory if CUDA is used.

    Returns
    -------
    torch.Tensor
        Output tensor.
    """
    assert (
        mat_a_tiles.shape[-1] == mat_b_tiles.shape[-2]
        and len(mat_a_tiles.shape) == 3
        and len(mat_b_tiles.shape) == 3
        and mat_a_tiles.shape[-2] != 0
    ), "Incompatible tile shapes used."
    if source_resistance is not None and line_resistance is not None:
        assert (
            source_resistance != 0 or line_resistance != 0
        ), "R_source or R_line must be non-zero."
        assert (
            source_resistance >= 0 and line_resistance >= 0
        ), "R_source and R_line must be >=0."

    if use_bindings:
        if quant_method is None:
            return memtorch_bindings.tile_matmul(
                mat_a_tiles,
                mat_a_tiles_map,
                mat_a_shape,
                mat_b_tiles,
                mat_b_tiles_map,
                mat_b_shape,
                cuda_malloc_heap_size,
            )
        else:
            assert (
                quant_method in memtorch.bh.Quantize.quant_methods
            ), "quant_method is invalid."
            return memtorch_bindings.tile_matmul(
                mat_a_tiles,
                mat_a_tiles_map,
                mat_a_shape,
                mat_b_tiles,
                mat_b_tiles_map,
                mat_b_shape,
                ADC_resolution,
                ADC_overflow_rate,
                memtorch.bh.Quantize.quant_methods.index(quant_method),
                cuda_malloc_heap_size,
            )
    else:
        result = torch.zeros((mat_a_shape[0], mat_b_shape[1]))
        if mat_a_tiles.shape[-2] > 1:
            for row_idx in range(mat_a_tiles.shape[-2]):
                result[row_idx] = tile_matmul_row(
                    mat_a_tiles[:, row_idx, :],
                    mat_a_tiles_map,
                    mat_b_tiles,
                    mat_b_tiles_map,
                    mat_b_shape,
                    source_resistance,
                    line_resistance,
                    ADC_resolution,
                    ADC_overflow_rate,
                    quant_method,
                    transistor,
                )
        else:
            result = tile_matmul_row(
                mat_a_tiles,
                mat_a_tiles_map,
                mat_b_tiles,
                mat_b_tiles_map,
                mat_b_shape,
                source_resistance,
                line_resistance,
                ADC_resolution,
                ADC_overflow_rate,
                quant_method,
                transistor,
            )
        return result


def tiled_inference(input, m, transistor):
    """Method to perform tiled inference.

    Parameters
    ----------
    input : torch.Tensor
        Input tensor (2-D).
    m : memtorch.mn
        Memristive MemTorch layer.

    Returns
    -------
    torch.Tensor
        Output tensor.
    """
    tiles_map = m.crossbars[0].tiles_map
    crossbar_shape = (m.crossbars[0].rows, m.crossbars[0].columns)
    if m.source_resistance is not None and m.line_resistance is not None:
        assert (
            m.source_resistance != 0 or m.line_resistance != 0
        ), "R_source or R_line must be non-zero."
        assert (
            m.source_resistance >= 0 and m.line_resistance >= 0
        ), "R_source and R_line must be >=0."

    if m.use_bindings:
        quant_method = m.quant_method
        if quant_method is None:
            if transistor:
                if "cpu" in memtorch.__version__:
                    return memtorch_bindings.tiled_inference(
                        input,
                        input.shape,
                        m.tile_shape,
                        m.crossbar_operation(
                            m.crossbars, lambda crossbar: crossbar.conductance_matrix
                        ),
                        m.crossbars[0].tiles_map,
                        (m.crossbars[0].rows, m.crossbars[0].columns),
                    )
                else:
                    return memtorch_bindings.tiled_inference(
                        input,
                        input.shape,
                        m.tile_shape,
                        m.crossbar_operation(
                            m.crossbars, lambda crossbar: crossbar.conductance_matrix
                        ),
                        m.crossbars[0].tiles_map,
                        (m.crossbars[0].rows, m.crossbars[0].columns),
                        cuda_malloc_heap_size=m.cuda_malloc_heap_size,
                    )
            else:
                if "cpu" in memtorch.__version__:
                    return memtorch_bindings.tiled_inference(
                        input,
                        input.shape,
                        m.tile_shape,
                        m.crossbar_operation(
                            m.crossbars, lambda crossbar: crossbar.conductance_matrix
                        ),
                        m.crossbars[0].tiles_map,
                        (m.crossbars[0].rows, m.crossbars[0].columns),
                        m.source_resistance,
                        m.line_resistance,
                    )
                else:
                    return memtorch_bindings.tiled_inference(
                        input,
                        input.shape,
                        m.tile_shape,
                        m.crossbar_operation(
                            m.crossbars, lambda crossbar: crossbar.conductance_matrix
                        ),
                        m.crossbars[0].tiles_map,
                        (m.crossbars[0].rows, m.crossbars[0].columns),
                        m.source_resistance,
                        m.line_resistance,
                        cuda_malloc_heap_size=m.cuda_malloc_heap_size,
                    )
        else:
            assert (
                quant_method in memtorch.bh.Quantize.quant_methods
            ), "quant_method is invalid."
            if transistor:
                if "cpu" in memtorch.__version__:
                    return memtorch_bindings.tiled_inference(
                        input,
                        input.shape,
                        m.tile_shape,
                        m.crossbar_operation(
                            m.crossbars, lambda crossbar: crossbar.conductance_matrix
                        ),
                        tiles_map,
                        crossbar_shape,
                        m.ADC_resolution,
                        m.ADC_overflow_rate,
                        memtorch.bh.Quantize.quant_methods.index(quant_method),
                    )
                else:
                    return memtorch_bindings.tiled_inference(
                        input,
                        input.shape,
                        m.tile_shape,
                        m.crossbar_operation(
                            m.crossbars, lambda crossbar: crossbar.conductance_matrix
                        ),
                        tiles_map,
                        crossbar_shape,
                        m.ADC_resolution,
                        m.ADC_overflow_rate,
                        memtorch.bh.Quantize.quant_methods.index(quant_method),
                        cuda_malloc_heap_size=m.cuda_malloc_heap_size,
                    )
            else:
                if "cpu" in memtorch.__version__:
                    return memtorch_bindings.tiled_inference(
                        input,
                        input.shape,
                        m.tile_shape,
                        m.crossbar_operation(
                            m.crossbars, lambda crossbar: crossbar.conductance_matrix
                        ),
                        tiles_map,
                        crossbar_shape,
                        m.source_resistance,
                        m.line_resistance,
                        m.ADC_resolution,
                        m.ADC_overflow_rate,
                        memtorch.bh.Quantize.quant_methods.index(quant_method),
                    )
                else:
                    return memtorch_bindings.tiled_inference(
                        input,
                        input.shape,
                        m.tile_shape,
                        m.crossbar_operation(
                            m.crossbars, lambda crossbar: crossbar.conductance_matrix
                        ),
                        tiles_map,
                        crossbar_shape,
                        m.source_resistance,
                        m.line_resistance,
                        m.ADC_resolution,
                        m.ADC_overflow_rate,
                        memtorch.bh.Quantize.quant_methods.index(quant_method),
                        cuda_malloc_heap_size=m.cuda_malloc_heap_size,
                    )
    else:
        (input_tiles, input_tiles_map) = gen_tiles(
            input,
            m.tile_shape,
            input=True,
            use_bindings=False,
        )
        return tile_matmul(
            input_tiles,
            input_tiles_map,
            input.shape,
            m.crossbar_operation(
                m.crossbars, lambda crossbar: crossbar.conductance_matrix
            ),
            tiles_map,
            crossbar_shape,
            m.source_resistance,
            m.line_resistance,
            m.ADC_resolution,
            m.ADC_overflow_rate,
            m.quant_method,
            m.transistor,
            use_bindings=False,
        )



"""
Empirical Metal-Oxide RRAM Device Endurance and Retention Model as described in:
C. Lammie, M. Rahimi Azghadi and D. Ielmini, "Empirical Metal-Oxide RRAM Device Endurance and Retention Model For Deep Learning Simulations", Semiconductor Science and Technology, 2021.
"""
import enum
import importlib
import math
from enum import Enum, auto

import numpy as np
import scipy
import torch
from scipy.interpolate import interp1d


class OperationMode(Enum):
    sudden = auto()
    gradual = auto()


def scale_p_0(p_0, p_1, v_stop, v_stop_min, v_stop_max, v_stop_optimal, cell_size=10):
    """Method to scale p_0 to introduce dependence on v_stop

    Parameters
    ----------
    p_0 : float
        Unscaled p_0 parameter.
    p_1 : float
        p_1 parameter.
    v_stop : float
        v_stop value to determine p_0 for.
    v_stop_min : float
        minimum v_stop value.
    v_stop_max : float
        maximum v_stop value.
    v_stop_optimal : float
        optimal v_stop value (endurance).
    cell_size : float
        Device cell size (nm).

    Returns
    -------
    float
        Scaled p_0 value.
    """
    assert (
        v_stop_max > v_stop_min
        and v_stop_optimal >= v_stop_min
        and v_stop_optimal <= v_stop_max
    ), "v_stop_max must be larger than v_stop_min and v_stop_optimal must be >= v_stop_min and <= v_stop_max."
    if cell_size is None:
        cell_size = 10

    scale_input = interp1d([v_stop_min, v_stop_max], [0, 1])
    scaled_input = scale_input(v_stop)
    y = p_0 * np.exp(p_1 * cell_size)
    k = np.log10(y) / (1 - (2 * scale_input(v_stop_optimal) - 1) ** (2))
    return (10 ** (k * (1 - (2 * scaled_input - 1) ** (2)))) / (np.exp(p_1 * cell_size))


def model_endurance_retention_gradual(
    initial_resistance,
    x,
    p_0,
    p_1,
    p_2,
    p_3,
    threshold,
    temperature_constant,
    cell_size=None,
):
    """Method to model gradual endurance_retention failure.

    Parameters
    ----------
    initial_resistance : tensor
        Initial resistance values.
     x : float
        Energy (J) / SET-RESET cycles / Retention time (s).
    p_0 : float
        p_0 fitting parameter.
    p_1 : float
        p_1 fitting parameter.
    p_2 : float
        p_2 fitting parameter.
    p_3 : float
        p_3 fitting parameter.
    threshold : float
        Threshold for x.
    temperature_threshold : float
        Temperature threshold (K) in which the device begins to fail.
    cell_size : float
        Device cell size (nm).

    Returns
    -------
    tensor
        Updated resistance values.
    """
    return 10 ** (
        p_3 * (p_1 * cell_size + p_2 * temperature_constant) * np.log10(x)
        + torch.log10(initial_resistance)
        - p_3 * (p_1 * cell_size + p_2 * temperature_constant) * np.log10(threshold)
    )


def model_endurance_retention(
    layer,
    operation_mode,
    x,
    p_lrs,
    stable_resistance_lrs,
    p_hrs,
    stable_resistance_hrs,
    cell_size,
    temperature,
    temperature_threshold=298,
):
    """Method to model endurance and retention characteristics.

    Parameters
    ----------
    layer : memtorch.mn
        A memrstive layer.
    operation_mode: memtorch.bh.endurance_retention_models.OperationMode
        Failure operational mode (sudden or gradual).
    x : float
        x parameter.
    p_lrs : list
        Low resistance state p_0, p_1, p_2, and p_3 values.
    stable_resistance_lrs : float
        Stable low resistance state.
    p_hrs : list
        High resistance state p_0, p_1, p_2, and p_3 values.
    stable_resistance_hrs : float
        Stable high resistance state.
    cell_size : float
        Device cell size (nm).
    temperature : float
        Operational temperature (K).
    temperature_threshold : float
        Temperature threshold (K) in which the device begins to fail.

    Returns
    -------
    memtorch.mn
        The patched memristive layer.
    """
    assert (len(p_lrs) == 4 or p_lrs is None) and (
        len(p_hrs) == 4 or p_hrs is None
    ), "p_lrs or p_hrs are of invalid length."
    if cell_size is None:
        cell_size = 10

    if temperature is None or temperature_threshold is None:
        temperature_constant = 0
    else:
        temperature_constant = min(temperature_threshold / temperature, 1)

    if p_lrs is not None:
        if temperature is None:
            threshold_lrs = p_lrs[0] * np.exp(p_lrs[1])
        else:
            threshold_lrs = p_lrs[0] * np.exp(
                p_lrs[1] * cell_size + p_lrs[2] * temperature_constant
            )

    if p_hrs is not None:
        if temperature is None:
            threshold_hrs = p_hrs[0] * np.exp(p_hrs[1])
        else:
            threshold_hrs = p_hrs[0] * np.exp(
                p_hrs[1] * cell_size + p_hrs[2] * temperature_constant
            )

    for i in range(len(layer.crossbars)):
        initial_resistance = 1 / layer.crossbars[i].conductance_matrix
        convergence_point = (
            layer.crossbars[i].r_on_mean + layer.crossbars[i].r_off_mean / 2
        )
        if x > threshold_lrs:
            if (
                operation_mode == OperationMode.gradual
                and p_lrs is not None
                and initial_resistance[
                    initial_resistance < convergence_point
                ].nelement()
                > 0
            ):
                initial_resistance[
                    initial_resistance < convergence_point
                ] = model_endurance_retention_gradual(
                    initial_resistance=initial_resistance[
                        initial_resistance < convergence_point
                    ],
                    x=x,
                    p_0=p_lrs[0],
                    p_1=p_lrs[1],
                    p_2=p_lrs[2],
                    p_3=p_lrs[3],
                    threshold=threshold_lrs,
                    temperature_constant=temperature_constant,
                    cell_size=cell_size,
                )
            elif operation_mode == OperationMode.sudden:
                initial_resistance[
                    initial_resistance < convergence_point
                ] = stable_resistance_lrs

        if x > threshold_hrs:
            if (
                operation_mode == OperationMode.gradual
                and p_hrs is not None
                and initial_resistance[
                    initial_resistance > convergence_point
                ].nelement()
                > 0
            ):
                initial_resistance[
                    initial_resistance > convergence_point
                ] = model_endurance_retention_gradual(
                    initial_resistance=initial_resistance[
                        initial_resistance > convergence_point
                    ],
                    x=x,
                    p_0=p_hrs[0],
                    p_1=p_hrs[1],
                    p_2=p_hrs[2],
                    p_3=p_hrs[3],
                    threshold=threshold_hrs,
                    temperature_constant=temperature_constant,
                    cell_size=cell_size,
                )
            elif operation_mode == OperationMode.sudden and x > threshold_hrs:
                initial_resistance[
                    initial_resistance > convergence_point
                ] = stable_resistance_hrs

        layer.crossbars[i].conductance_matrix = 1 / initial_resistance

    return layer



import math

import numpy as np
import torch

import memtorch


def Prodromakis(x, p=1, j=1.0):
    """Prodromakis window function.

    Parameters
    ----------
    x : float
        State variable.
    p : int
        p constant.
    j : float
        j constant.
    """
    return j * ((1 - (x - 0.5) ** 2 + 0.75) ** p)



import copy
import math
import random

import numpy as np
import pytest
import torch

import memtorch
from memtorch.bh.crossbar.Program import naive_program
from memtorch.map.Parameter import naive_map
from memtorch.mn.Module import patch_model


@pytest.mark.parametrize("mean, std", [(0, 2.5), (0, 5)])
def test_stochastic_parameter(mean, std):
    with pytest.raises(Exception):
        stochastic_parameter = memtorch.bh.StochasticParameter(invalid_arg=None)

    stochastic_parameter = memtorch.bh.StochasticParameter(loc=mean, scale=std)
    assert type(stochastic_parameter(return_mean=True).item()) == float

    class TestObject:
        def __init__(self, test_parameter):
            args = memtorch.bh.unpack_parameters(locals())
            self.test_parameter = args.test_parameter

    kwargs = {"test_parameter": stochastic_parameter}
    test_object = TestObject(**kwargs)
    assert isinstance(
        test_object.test_parameter, (int, float, complex)
    ) and not isinstance(test_object.test_parameter, bool)
    parameter = random.random()
    kwargs = {"test_parameter": parameter}
    test_object = TestObject(**kwargs)
    assert isinstance(
        test_object.test_parameter, (int, float, complex)
    ) and not isinstance(test_object.test_parameter, bool)
    assert (
        type(memtorch.bh.StochasticParameter(loc=mean, scale=std, function=False))
        == float
    )


def test_resample_r_off_r_on(debug_networks):
    networks = debug_networks
    for network in networks:
        with pytest.raises(Exception):
            patched_network = patch_model(
                copy.deepcopy(network),
                memristor_model=memtorch.bh.memristor.LinearIonDrift,
                memristor_model_params={
                    "r_off": memtorch.bh.StochasticParameter(loc=1, scale=0),
                    "r_on": memtorch.bh.StochasticParameter(loc=1, scale=0),
                },
                module_parameters_to_patch=[type(network.layer)],
                mapping_routine=naive_map,
                transistor=True,
                programming_routine=None,
                scheme=memtorch.bh.Scheme.SingleColumn,
            )
        with pytest.raises(Exception):
            patched_network = patch_model(
                copy.deepcopy(network),
                memristor_model=memtorch.bh.memristor.LinearIonDrift,
                memristor_model_params={"r_off": 1, "r_on": 1},
                module_parameters_to_patch=[type(network.layer)],
                mapping_routine=naive_map,
                transistor=True,
                programming_routine=None,
                scheme=memtorch.bh.Scheme.SingleColumn,
            )
